# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""API for providers."""

import abc
from collections import UserDict, defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Protocol, Tuple, Type, Union

from renku.core import errors
from renku.core.constant import ProviderPriority
from renku.core.plugin import hookimpl
from renku.core.util import communication
from renku.core.util.os import delete_dataset_file
from renku.core.util.urls import is_uri_subfolder, resolve_uri
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.domain_model.dataset import RemoteEntity
from renku.domain_model.dataset_provider import IDatasetProviderPlugin
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import (
        DatasetAddMetadata,
        DatasetUpdateMetadata,
        ProviderDataset,
        ProviderDatasetFile,
        ProviderParameter,
    )
    from renku.core.interface.storage import IStorage
    from renku.domain_model.dataset import Dataset, DatasetTag
    from renku.infrastructure.immutable import DynamicProxy


class ProviderApi(IDatasetProviderPlugin):
    """Interface defining provider methods."""

    priority: Optional[ProviderPriority] = None
    name: Optional[str] = None
    is_remote: Optional[bool] = None

    def __init__(self, uri: str, **kwargs):
        self._uri: str = uri or ""

    def __init_subclass__(cls, **kwargs):
        for required_property in ("priority", "name", "is_remote"):
            if getattr(cls, required_property, None) is None:
                raise NotImplementedError(f"{required_property} must be set for {cls}")

    def __repr__(self):
        return f"<DataProvider {self.name}>"

    @classmethod
    @hookimpl
    def dataset_provider(cls) -> "Type[ProviderApi]":
        """The definition of the provider."""
        return cls

    @staticmethod
    @abc.abstractmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        raise NotImplementedError

    @property
    def uri(self) -> str:
        """Return provider's URI."""
        return self._uri


class AddProviderInterface(abc.ABC):
    """Interface defining providers that can add data to a dataset."""

    @staticmethod
    def get_add_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for add."""
        return []

    @abc.abstractmethod
    def get_metadata(self, uri: str, destination: Path, **kwargs) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        raise NotImplementedError

    @abc.abstractmethod
    def update_files(
        self, files: List["DynamicProxy"], dry_run: bool, delete: bool, context: Dict[str, Any], **kwargs
    ) -> List["DatasetUpdateMetadata"]:
        """Update dataset files from the remote provider."""
        raise NotImplementedError


class ExportProviderInterface(abc.ABC):
    """Interface defining export providers."""

    @staticmethod
    def get_export_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for export."""
        return []

    @abc.abstractmethod
    def get_exporter(self, dataset: "Dataset", *, tag: Optional["DatasetTag"], **kwargs) -> "ExporterApi":
        """Get export manager."""
        raise NotImplementedError


class ImportProviderInterface(abc.ABC):
    """Interface defining import providers."""

    @staticmethod
    def get_import_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for import."""
        return []

    @abc.abstractmethod
    def get_importer(self, **kwargs) -> "ImporterApi":
        """Get import manager."""
        raise NotImplementedError


class StorageProviderInterface(abc.ABC):
    """Interface defining backend storage providers."""

    @abc.abstractmethod
    def get_credentials(self) -> "ProviderCredentials":
        """Return an instance of provider's credential class."""
        raise NotImplementedError

    @abc.abstractmethod
    def convert_to_storage_uri(self, uri: str) -> str:
        """Convert backend-specific URI to a URI that is usable by the IStorage implementation."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_storage(self, credentials: Optional["ProviderCredentials"] = None) -> "IStorage":
        """Return the storage manager for the provider."""
        raise NotImplementedError

    @abc.abstractmethod
    def on_create(self, dataset: "Dataset") -> None:
        """Hook to perform provider-specific actions on a newly-created dataset."""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def supports_storage(uri: str) -> bool:
        """Whether or not this provider supports a given URI storage."""
        raise NotImplementedError

    def update_files(
        self,
        files: List["DynamicProxy"],
        dry_run: bool,
        delete: bool,
        context: Dict[str, Any],
        **kwargs,
    ) -> List["DatasetUpdateMetadata"]:
        """Update dataset files from the remote provider."""
        from renku.core.dataset.providers.models import DatasetUpdateAction, DatasetUpdateMetadata

        progress_text = f"Checking remote files for updates in dataset {files[0].dataset.slug}"

        results: List[DatasetUpdateMetadata] = []

        try:
            communication.start_progress(progress_text, len(files))

            storage = self.get_storage()

            # group files by storage to efficiently compute hashes
            storage_files_dict: Dict[str, List["DynamicProxy"]] = defaultdict(list)

            for file in files:
                if file.dataset.storage:
                    storage_files_dict[file.dataset.storage].append(file)
                elif file.based_on:
                    if not self.supports_storage(file.based_on.url):
                        raise ValueError(
                            f"Called {getattr(self, 'name', 'Storage')} provider with file {file.entity.path} "
                            "which is not supported by this provider"
                        )
                    storage_files_dict[file.based_on.url].append(file)

            for file_storage, files in storage_files_dict.items():
                hashes = storage.get_hashes(uri=file_storage)
                for file in files:
                    communication.update_progress(progress_text, 1)
                    if not file.based_on:
                        continue

                    dst = project_context.metadata_path.parent / file.entity.path

                    hash = next((h for h in hashes if h.uri == file.based_on.url), None)

                    if hash:
                        if not dry_run and (
                            not file.dataset.storage
                            or not is_uri_subfolder(resolve_uri(file.dataset.storage), file.based_on.url)
                        ):
                            # Redownload downloaded (not mounted) file
                            download_storage = self.get_storage()
                            download_storage.download(file.based_on.url, dst)
                        file.based_on = RemoteEntity(
                            checksum=hash.hash if hash.hash else "", url=hash.uri, path=hash.path
                        )
                        results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.UPDATE))
                    else:
                        if (
                            not dry_run
                            and delete
                            and (
                                not file.dataset.storage
                                or not is_uri_subfolder(resolve_uri(file.dataset.storage), file.based_on.url)
                            )
                        ):
                            # Delete downloaded (not mounted) file
                            delete_dataset_file(dst, follow_symlinks=True)
                            project_context.repository.add(dst, force=True)
                        results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))

        finally:
            communication.finalize_progress(progress_text)

        return results


class CloudStorageProviderType(Protocol):
    """Intersection type for ``mypy`` hinting in storage classes."""

    @property
    def uri(self) -> str:
        """Return provider's URI."""
        raise NotImplementedError

    @abc.abstractmethod
    def convert_to_storage_uri(self, uri: str) -> str:
        """Convert backend-specific URI to a URI that is usable by the IStorage implementation."""
        raise NotImplementedError


class ImporterApi(abc.ABC):
    """Interface defining importer methods."""

    def __init__(self, uri: str, original_uri: str):
        self._uri: str = uri
        self._original_uri: str = original_uri
        self._provider_dataset_files: Optional[List["ProviderDatasetFile"]] = None
        self._provider_dataset: Optional["ProviderDataset"] = None

    @property
    def provider_dataset(self) -> "ProviderDataset":
        """Return the remote dataset. This is only valid after a call to ``fetch_provider_dataset``."""
        if self._provider_dataset is None:
            raise errors.DatasetImportError("Dataset is not fetched")

        return self._provider_dataset

    @property
    def provider_dataset_files(self) -> List["ProviderDatasetFile"]:
        """Return list of dataset files. This is only valid after a call to ``fetch_provider_dataset``."""
        if self._provider_dataset_files is None:
            raise errors.DatasetImportError("Dataset is not fetched")

        return self._provider_dataset_files

    @property
    def uri(self) -> str:
        """Return url of this record."""
        return self._uri

    @property
    def original_uri(self) -> str:
        """Return original URI of this record without any conversion to DOI."""
        return self._original_uri

    @property
    def latest_uri(self) -> str:
        """Get URI of the latest version."""
        raise NotImplementedError

    @property
    def version(self) -> str:
        """Get record version."""
        raise NotImplementedError

    @abc.abstractmethod
    def fetch_provider_dataset(self) -> "ProviderDataset":
        """Deserialize this record to a ``ProviderDataset``."""
        raise NotImplementedError

    @abc.abstractmethod
    def is_latest_version(self) -> bool:
        """Check if record is at last possible version."""
        raise NotImplementedError

    def is_version_equal_to(self, dataset: Any) -> bool:
        """Check if a dataset has the same version as the record."""
        return self.version == getattr(dataset, "version", object())

    @abc.abstractmethod
    def download_files(self, destination: Path, extract: bool) -> List["DatasetAddMetadata"]:
        """Download dataset files from the remote provider."""
        raise NotImplementedError

    @abc.abstractmethod
    def tag_dataset(self, name: str) -> None:
        """Create a tag for the dataset ``name`` if the remote dataset has a tag/version."""
        raise NotImplementedError

    @abc.abstractmethod
    def copy_extra_metadata(self, new_dataset: "Dataset") -> None:
        """Copy provider specific metadata once the dataset is created."""
        raise NotImplementedError


class ExporterApi(abc.ABC):
    """Interface defining exporter methods."""

    def __init__(self, dataset: "Dataset"):
        super().__init__()
        self._dataset: "Dataset" = dataset

    @property
    def dataset(self) -> "Dataset":
        """Return the dataset to be exported."""
        return self._dataset

    @staticmethod
    def requires_access_token() -> bool:
        """Return if export requires an access token."""
        return True

    @abc.abstractmethod
    def set_access_token(self, access_token):
        """Set access token."""
        pass

    @abc.abstractmethod
    def get_access_token_url(self) -> str:
        """Endpoint for creation of access token."""
        pass

    @abc.abstractmethod
    def export(self, **kwargs) -> str:
        """Execute export process."""
        raise NotImplementedError


class ProviderCredentials(abc.ABC, UserDict):
    """Credentials of a provider.

    NOTE: An empty string, "", is a valid value. ``NO_VALUE`` means that the value for a key is not set.
    """

    def __init__(self, provider: ProviderApi):
        super().__init__()
        self._provider: ProviderApi = provider
        self.data: Dict[str, Union[str, NoValueType]] = {
            key: NO_VALUE for key in self.get_canonical_credentials_names()
        }

    @staticmethod
    @abc.abstractmethod
    def get_credentials_names() -> Tuple[str, ...]:
        """Return a tuple of the required credentials for a provider."""
        raise NotImplementedError

    @property
    def provider(self):
        """Return the associated provider instance."""
        return self._provider

    def get_credentials_names_with_no_value(self) -> Tuple[str, ...]:
        """Return a tuple of credential keys that don't have a valid value."""
        return tuple(key for key, value in self.items() if value is NO_VALUE)

    def get_canonical_credentials_names(self) -> Tuple[str, ...]:
        """Return canonical credentials names that can be used as config keys."""
        from renku.core.util.metadata import get_canonical_key

        return tuple(get_canonical_key(key) for key in self.get_credentials_names())

    def get_canonical_credentials_names_with_no_value(self) -> Tuple[str, ...]:
        """Return canonical credentials names that can be used as config keys for keys with no valid value."""
        from renku.core.util.metadata import get_canonical_key

        return tuple(get_canonical_key(key) for key in self.get_credentials_names_with_no_value())

    def get_credentials_section_name(self) -> str:
        """Get section name for storing credentials.

        NOTE: This methods should be overridden by subclasses to allow multiple credentials per providers if needed.
        NOTE: Values used in this method shouldn't depend on ProviderCredentials attributes since we don't have those
        attributes when reading credentials. It's OK to use ProviderApi attributes.
        """
        return self.provider.name.lower()  # type: ignore

    def read(self) -> Dict[str, Union[str, NoValueType]]:
        """Read credentials from the config and return them. Set non-existing values to None."""
        from renku.core.util.metadata import read_credentials

        section = self.get_credentials_section_name()

        def read_and_convert_credentials(key) -> Union[str, NoValueType]:
            value = read_credentials(section=section, key=key)
            return NO_VALUE if value is None else value

        data = {key: read_and_convert_credentials(key) for key in self.get_canonical_credentials_names_with_no_value()}
        self.data.update(data)

        return self.data

    def store(self) -> None:
        """Store credentials globally."""
        from renku.core.util.metadata import store_credentials

        section = self.get_credentials_section_name()

        for key, value in self.items():
            if value is not None:
                store_credentials(section=section, key=key, value=value)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service controller mixin."""
import contextlib
from abc import ABCMeta, abstractmethod
from functools import wraps
from pathlib import Path
from typing import Optional, Union

import portalocker

from renku.core.constant import RENKU_HOME
from renku.core.errors import LockError, RenkuException, UninitializedProject
from renku.core.util.contexts import renku_project_context
from renku.infrastructure.repository import Repository
from renku.ui.service.cache.config import REDIS_NAMESPACE
from renku.ui.service.cache.models.job import Job
from renku.ui.service.cache.models.project import Project
from renku.ui.service.cache.models.user import User
from renku.ui.service.config import PROJECT_CLONE_DEPTH_DEFAULT
from renku.ui.service.controllers.utils.remote_project import RemoteProject
from renku.ui.service.errors import (
    IntermittentAuthenticationError,
    IntermittentLockError,
    ProgramRenkuError,
    UserAnonymousError,
)
from renku.ui.service.gateways.repository_cache import LocalRepositoryCache
from renku.ui.service.jobs.contexts import enqueue_retry
from renku.ui.service.jobs.delayed_ctrl import delayed_ctrl_job
from renku.ui.service.serializers.common import DelayedResponseRPC
from renku.ui.service.utils import normalize_git_url

PROJECT_FETCH_TIME = 30


def local_identity(method):
    """Ensure identity on local execution."""

    @wraps(method)
    def _impl(self, *method_args, **method_kwargs):
        """Implementation of method wrapper."""
        if not hasattr(self, "user") and not isinstance(getattr(self, "user", None), User):
            raise UserAnonymousError()

        return method(self, *method_args, **method_kwargs)

    return _impl


class RenkuOperationMixin(metaclass=ABCMeta):
    """Renku operation execution mixin.

    It's responsible for setting up all dependencies for an execution of a Renku operation. This includes cloning of
    project repository, syncing with remote state, delaying write operations and returning job identifier.
    """

    JOB_RESPONSE_SERIALIZER = DelayedResponseRPC()

    def __init__(
        self,
        cache,
        user_data,
        request_data,
        migrate_project=False,
        skip_lock=False,
        clone_depth=PROJECT_CLONE_DEPTH_DEFAULT,
    ):
        """Read operation mixin for controllers."""
        if user_data and "user_id" in user_data and cache is not None:
            self.user = cache.ensure_user(user_data)

        self.is_write = False
        self.migrate_project = migrate_project
        self.skip_lock = skip_lock
        self.clone_depth = clone_depth
        self.cache = cache
        self.user_data = user_data
        self.request_data = request_data

        # NOTE: Manually set `migrate_project` flag takes higher precedence
        # than `migrate_project` flag from the request. Reason for this is
        # because we reuse view controllers in the delayed renku operation jobs.
        if not self.migrate_project and isinstance(self.request_data, dict):
            self.migrate_project = self.request_data.get("migrate_project", False)

        # NOTE: This is absolute project path and its set before invocation of `renku_op`,
        # so it's safe to use it in controller operations. Its type will always be `pathlib.Path`.
        self._project_path = None

    @property
    @abstractmethod
    def context(self):
        """Operation context."""
        raise NotImplementedError

    @property
    def project_path(self) -> Optional[Path]:
        """Absolute project's path."""
        return self._project_path

    @project_path.setter
    def project_path(self, path: Optional[Union[str, Path]]):
        """Set absolute project's path."""
        if not path:
            self._project_path = None
            return

        path = normalize_git_url(str(path))
        self._project_path = Path(path)

    @abstractmethod
    def renku_op(self):
        """Implements operation for the controller."""
        raise NotImplementedError

    def ensure_migrated(self, project: Project):
        """Ensure that project is migrated."""
        if not self.migrate_project:
            return

        from renku.ui.service.controllers.cache_migrate_project import MigrateProjectCtrl

        migrate_context = {
            "git_url": project.git_url,
            "skip_docker_update": True,
            "skip_template_update": True,
            "branch": project.branch,
        }
        migration_response = MigrateProjectCtrl(
            self.cache, self.user_data, migrate_context, skip_lock=True
        ).to_response()

        return migration_response

    def execute_op(self):
        """Execute renku operation which controller implements."""
        ctrl_cls = {
            "renku_module": self.__class__.__module__,
            "renku_ctrl": self.__class__.__name__,
        }

        if self.context.get("is_delayed", False) and "user_id" in self.user_data:
            # NOTE: After pushing the controller to delayed execution,
            # it's important to remove the delayed mark,
            # otherwise job will keep recursively enqueuing itself.
            self.context.pop("is_delayed")

            job = self.cache.make_job(self.user, job_data={"ctrl_context": {**self.context, **ctrl_cls}})

            with enqueue_retry(f"{REDIS_NAMESPACE}.delayed.ctrl.{ctrl_cls['renku_ctrl']}") as queue:
                queue.enqueue(delayed_ctrl_job, self.context, self.user_data, job.job_id, **ctrl_cls)

            return job

        if "git_url" in self.context:
            if "user_id" not in self.user_data:
                # NOTE: Anonymous session support.
                return self.remote()
            else:
                return self.local()
        else:
            raise RenkuException("context does not contain `project_id` or `git_url`")

    @local_identity
    def local(self):
        """Execute renku operation against service cache."""
        if self.user is None or self.cache is None:
            error = Exception("local execution is disabled")
            raise ProgramRenkuError(error)

        project = LocalRepositoryCache().get(
            self.cache,
            self.request_data["git_url"],
            self.request_data.get("branch"),
            self.user,
            self.clone_depth is not None,
            self.request_data.get("commit_sha"),
        )

        self.context["project_id"] = project.project_id

        if self.skip_lock:
            lock = contextlib.suppress()
        elif self.is_write or self.migrate_project:
            lock = project.write_lock()
        else:
            lock = project.read_lock()
        try:
            with project.concurrency_lock():
                with lock:
                    # NOTE: Get up-to-date version of object
                    current_project = Project.load(project.project_id)
                    if self.migrate_project:
                        self.ensure_migrated(current_project)

                    self.project_path = current_project.abs_path

                    with renku_project_context(self.project_path):
                        return self.renku_op()
        except (portalocker.LockException, portalocker.AlreadyLocked, LockError) as e:
            raise IntermittentLockError() from e

    def remote(self):
        """Execute renku operation against remote project."""
        if self.user_data and "token" not in self.user_data:
            raise IntermittentAuthenticationError()

        project = RemoteProject(self.user_data, self.request_data)

        with project.remote() as path:
            self.project_path = Path(path)

            if not (self.project_path / RENKU_HOME).exists():
                raise UninitializedProject(self.project_path)
            with renku_project_context(self.project_path):
                return self.renku_op()


class RenkuOpSyncMixin(RenkuOperationMixin, metaclass=ABCMeta):
    """Sync operation mixin.

    Extension of `RenkuOperationMixin` responsible for syncing all write operation with the remote. In case sync fails,
    it will create a new branch and push newly created branch to the remote and return the branch name to the client.
    """

    def sync(self, remote="origin"):
        """Sync with remote."""
        from renku.core.util.git import push_changes

        if self.project_path is None:
            raise RenkuException("unable to sync with remote since no operation has been executed")

        with Repository(self.project_path) as repository:
            return push_changes(repository, remote=remote)

    def execute_and_sync(self, remote="origin"):
        """Execute operation which controller implements and sync with the remote."""
        # NOTE: This will return the operation result as well as name of the branch to which it has been pushed.
        self.is_write = True

        result = self.execute_op()

        if isinstance(result, Job):
            return result, None

        if hasattr(result, "output"):
            result = result.output

        return result, self.sync(remote=remote)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON-LD dataset migrations."""

import itertools
import json
import os
import uuid
from pathlib import Path
from typing import Dict

import pyld

from renku.core.migration.utils import (
    OLD_METADATA_PATH,
    get_datasets_path,
    get_pre_0_3_4_datasets_metadata,
    is_using_temporary_datasets_path,
)
from renku.core.util.yaml import read_yaml, write_yaml
from renku.domain_model.project_context import project_context


def migrate(_):
    """Migration function."""

    _migrate_project_metadata()
    _migrate_datasets_metadata()


def _migrate_project_metadata():
    """Apply all initial JSON-LD migrations to project."""
    jsonld_translate = {
        "http://schema.org/name": "http://xmlns.com/foaf/0.1/name",
        "http://schema.org/Project": "http://xmlns.com/foaf/0.1/Project",
    }

    if project_context.metadata_path.joinpath(OLD_METADATA_PATH).exists():
        _apply_on_the_fly_jsonld_migrations(
            path=project_context.metadata_path.joinpath(OLD_METADATA_PATH),
            jsonld_context=_INITIAL_JSONLD_PROJECT_CONTEXT,
            fields=_PROJECT_FIELDS,
            jsonld_translate=jsonld_translate,
            persist_changes=not is_using_temporary_datasets_path(),
        )


def _migrate_datasets_metadata():
    """Apply all initial JSON-LD migrations to datasets."""
    jsonld_migrations = {
        "dctypes:Dataset": [_migrate_dataset_schema, _migrate_absolute_paths],
        "schema:Dataset": [
            _migrate_absolute_paths,
            _migrate_doi_identifier,
            _migrate_same_as_structure,
            _migrate_dataset_file_id,
        ],
    }

    old_metadata_paths = get_pre_0_3_4_datasets_metadata()
    new_metadata_paths = get_datasets_path().rglob(OLD_METADATA_PATH)

    for path in itertools.chain(old_metadata_paths, new_metadata_paths):
        _apply_on_the_fly_jsonld_migrations(
            path=path,
            jsonld_context=_INITIAL_JSONLD_DATASET_CONTEXT,
            fields=_DATASET_FIELDS,
            jsonld_migrations=jsonld_migrations,
        )


def _apply_on_the_fly_jsonld_migrations(
    path, jsonld_context, fields, jsonld_migrations=None, jsonld_translate=None, persist_changes=True
):
    data: Dict = read_yaml(path)

    if not isinstance(data, dict) and not isinstance(data, list):
        # NOTE: metadata file is probably not an actual renku file
        return

    if jsonld_translate:
        # perform the translation
        data = pyld.jsonld.expand(data)
        data_str = json.dumps(data)
        for k, v in jsonld_translate.items():
            data_str = data_str.replace(v, k)
        data = json.loads(data_str)
        data = pyld.jsonld.compact(data, jsonld_context)

    data.setdefault("@context", jsonld_context)

    _migrate_types(data)

    if jsonld_migrations:
        schema_type = data.get("@type")
        migrations = []

        if isinstance(schema_type, list):
            for schema in schema_type:
                migrations += jsonld_migrations.get(schema, [])
        elif isinstance(schema_type, str):
            migrations += jsonld_migrations.get(schema_type, [])

        for migration in set(migrations):
            data = migration(data)

    if data["@context"] != jsonld_context:
        # merge new context into old context to prevent properties
        # getting lost in jsonld expansion
        if isinstance(data["@context"], str):
            data["@context"] = {"@base": data["@context"]}
        data["@context"].update(jsonld_context)
        try:
            compacted = pyld.jsonld.compact(data, jsonld_context)
        except Exception:
            compacted = data
    else:
        compacted = data

    data = {}

    for k, v in compacted.items():
        if k in fields:
            no_value_context = isinstance(v, dict) and "@context" not in v
            has_nested_context = k in compacted["@context"] and "@context" in compacted["@context"][k]
            if no_value_context and has_nested_context:
                # Propagate down context
                v["@context"] = compacted["@context"][k]["@context"]

            data[k] = v

    data["@context"] = jsonld_context

    _migrate_types(data)

    if persist_changes:
        write_yaml(path, data)


def _migrate_dataset_schema(data):
    """Migrate from old dataset formats."""
    if "authors" not in data:
        return

    data["@context"]["creator"] = data["@context"].pop("authors", {"@container": "list"})

    data["creator"] = data.pop("authors", {})

    files = data.get("files", [])

    if isinstance(files, dict):
        files = files.values()
    for file_ in files:
        file_["creator"] = file_.pop("authors", {})

    return data


def _migrate_absolute_paths(data):
    """Migrate dataset paths to use relative path."""
    raw_path = data.get("path", ".")
    path = Path(raw_path)

    if path.is_absolute():
        try:
            data["path"] = str(path.relative_to(os.getcwd()))
        except ValueError:
            elements = raw_path.split("/")
            index = elements.index(".renku")
            data["path"] = str(Path("/".join(elements[index:])))

    files = data.get("files", [])

    if isinstance(files, dict):
        files = list(files.values())

    for file_ in files:
        path = Path(file_.get("path"), ".")
        if path.is_absolute():
            file_["path"] = str(path.relative_to(os.getcwd()))
    data["files"] = files
    return data


def _migrate_doi_identifier(data):
    """If the dataset _id is doi, make it a UUID."""
    from renku.core.util.doi import is_doi
    from renku.core.util.util import is_uuid

    _id = data.get("_id", "")
    identifier = data.get("identifier", "")

    if not is_uuid(_id):
        if not is_uuid(identifier):
            possible_identifier = Path(identifier).name
            if not is_uuid(possible_identifier):
                data["identifier"] = str(uuid.uuid4())
            else:
                data["identifier"] = possible_identifier
        if is_doi(data.get("_id", "")):
            data["same_as"] = {"@type": ["schema:URL"], "url": data["_id"]}
            if data.get("@context"):
                data["@context"].setdefault(
                    "same_as",
                    {
                        "@id": "schema:sameAs",
                        "@type": "schema:URL",
                        "@context": {"@version": "1.1", "url": "schema:url", "schema": "http://schema.org/"},
                    },
                )
        data["_id"] = data["identifier"]
    return data


def _migrate_same_as_structure(data):
    """Changes sameAs string to schema:URL object."""
    same_as = data.get("same_as")

    if same_as and isinstance(same_as, str):
        data["same_as"] = {"@type": ["schema:URL"], "url": same_as}

        if data.get("@context"):
            data["@context"].setdefault(
                "same_as",
                {
                    "@id": "schema:sameAs",
                    "@type": "schema:URL",
                    "@context": {"@version": "1.1", "url": "schema:url", "schema": "http://schema.org/"},
                },
            )

    return data


def _migrate_dataset_file_id(data):
    """Ensure dataset files have a fully qualified url as id."""
    host = "localhost"
    host = project_context.remote.host or host
    host = os.environ.get("RENKU_DOMAIN") or host

    files = data.get("files", [])
    for file_ in files:
        if not file_["_id"].startswith("http"):
            file_["_id"] = "https://{host}/{id}".format(host=host, id=file_["_id"])

    return data


def _migrate_types(data):
    """Fix types."""
    from renku.core.migration.utils import migrate_types

    migrate_types(data)


_PROJECT_FIELDS = {"_id", "created", "creator", "name", "updated", "version"}

_DATASET_FIELDS = {
    "_id",
    "_label",
    "_project",
    "based_on",
    "created",
    "creator",
    "date_published",
    "description",
    "files",
    "identifier",
    "in_language",
    "keywords",
    "license",
    "name",
    "path",
    "same_as",
    "short_name",
    "tags",
    "url",
    "version",
}

_INITIAL_JSONLD_PROJECT_CONTEXT = {
    "schema": "http://schema.org/",
    "prov": "http://www.w3.org/ns/prov#",
    "@version": 1.1,
    "name": "schema:name",
    "created": "schema:dateCreated",
    "updated": "schema:dateUpdated",
    "version": "schema:schemaVersion",
    "creator": {
        "@id": "schema:creator",
        "@context": {
            "schema": "http://schema.org/",
            "prov": "http://www.w3.org/ns/prov#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "@version": 1.1,
            "name": "schema:name",
            "email": "schema:email",
            "label": "rdfs:label",
            "affiliation": "schema:affiliation",
            "alternate_name": "schema:alternateName",
            "_id": "@id",
        },
    },
    "_id": "@id",
}

_INITIAL_JSONLD_DATASET_CONTEXT = {
    "schema": "http://schema.org/",
    "@version": 1.1,
    "prov": "http://www.w3.org/ns/prov#",
    "wfprov": "http://purl.org/wf4ever/wfprov#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "path": "prov:atLocation",
    "_id": "@id",
    "_project": {
        "@id": "schema:isPartOf",
        "@context": {
            "schema": "http://schema.org/",
            "prov": "http://www.w3.org/ns/prov#",
            "@version": 1.1,
            "name": "schema:name",
            "created": "schema:dateCreated",
            "updated": "schema:dateUpdated",
            "version": "schema:schemaVersion",
            "creator": {
                "@id": "schema:creator",
                "@context": {
                    "schema": "http://schema.org/",
                    "prov": "http://www.w3.org/ns/prov#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "@version": 1.1,
                    "name": "schema:name",
                    "email": "schema:email",
                    "label": "rdfs:label",
                    "affiliation": "schema:affiliation",
                    "alternate_name": "schema:alternateName",
                    "_id": "@id",
                },
            },
            "_id": "@id",
        },
    },
    "creator": {
        "@id": "schema:creator",
        "@context": {
            "schema": "http://schema.org/",
            "prov": "http://www.w3.org/ns/prov#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "@version": 1.1,
            "name": "schema:name",
            "email": "schema:email",
            "label": "rdfs:label",
            "affiliation": "schema:affiliation",
            "alternate_name": "schema:alternateName",
            "_id": "@id",
        },
    },
    "date_published": "schema:datePublished",
    "description": "schema:description",
    "identifier": "schema:identifier",
    "in_language": {
        "@id": "schema:inLanguage",
        "@context": {
            "schema": "http://schema.org/",
            "@version": 1.1,
            "alternate_name": "schema:alternateName",
            "name": "schema:name",
        },
    },
    "keywords": "schema:keywords",
    "based_on": "schema:isBasedOn",
    "license": "schema:license",
    "name": "schema:name",
    "url": "schema:url",
    "version": "schema:version",
    "created": "schema:dateCreated",
    "files": {
        "@id": "schema:hasPart",
        "@context": {
            "schema": "http://schema.org/",
            "@version": 1.1,
            "prov": "http://www.w3.org/ns/prov#",
            "wfprov": "http://purl.org/wf4ever/wfprov#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "path": "prov:atLocation",
            "_id": "@id",
            "_label": "rdfs:label",
            "_project": {
                "@id": "schema:isPartOf",
                "@context": {
                    "schema": "http://schema.org/",
                    "prov": "http://www.w3.org/ns/prov#",
                    "@version": 1.1,
                    "name": "schema:name",
                    "created": "schema:dateCreated",
                    "updated": "schema:dateUpdated",
                    "version": "schema:schemaVersion",
                    "creator": {
                        "@id": "schema:creator",
                        "@context": {
                            "schema": "http://schema.org/",
                            "prov": "http://www.w3.org/ns/prov#",
                            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                            "@version": 1.1,
                            "name": "schema:name",
                            "email": "schema:email",
                            "label": "rdfs:label",
                            "affiliation": "schema:affiliation",
                            "alternate_name": "schema:alternateName",
                            "_id": "@id",
                        },
                    },
                    "_id": "@id",
                },
            },
            "creator": {
                "@id": "schema:creator",
                "@context": {
                    "schema": "http://schema.org/",
                    "prov": "http://www.w3.org/ns/prov#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "@version": 1.1,
                    "name": "schema:name",
                    "email": "schema:email",
                    "label": "rdfs:label",
                    "affiliation": "schema:affiliation",
                    "alternate_name": "schema:alternateName",
                    "_id": "@id",
                },
            },
            "added": "schema:dateCreated",
            "name": "schema:name",
            "url": "schema:url",
            "external": "renku:external",
            "based_on": "schema:isBasedOn",
            "renku": "https://swissdatasciencecenter.github.io/renku-ontology#",
        },
    },
    "tags": {
        "@id": "schema:subjectOf",
        "@context": {
            "schema": "http://schema.org/",
            "@version": 1.1,
            "name": "schema:name",
            "description": "schema:description",
            "commit": "schema:location",
            "created": "schema:startDate",
            "dataset": "schema:about",
            "_id": "@id",
        },
    },
    "same_as": {
        "@id": "schema:sameAs",
        "@context": {"schema": "http://schema.org/", "@version": 1.1, "url": "schema:url", "_id": "@id"},
    },
    "short_name": "schema:alternateName",
}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache job related models."""
import uuid
from datetime import datetime

from marshmallow import fields, post_load

from renku.ui.service.cache.models.job import USER_JOB_STATE_ENQUEUED, Job
from renku.ui.service.serializers.common import CreationSchema, MandatoryUserSchema


class JobSchema(CreationSchema, MandatoryUserSchema):
    """Job serialization."""

    updated_at = fields.DateTime(load_default=datetime.utcnow)

    job_id = fields.String(load_default=lambda: uuid.uuid4().hex)
    project_id = fields.String()
    renku_op = fields.String()

    state = fields.String(required=False, load_default=USER_JOB_STATE_ENQUEUED)
    extras = fields.Dict(required=False)
    client_extras = fields.String(required=False, load_default=None)

    ctrl_result = fields.Dict(required=False)
    ctrl_context = fields.Dict(required=False)

    @post_load
    def make_job(self, data, **options):
        """Construct job object."""
        return Job(**data)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service error handlers."""

import re
from functools import wraps
from typing import cast

from jwt import ExpiredSignatureError, ImmatureSignatureError, InvalidIssuedAtError
from marshmallow import ValidationError
from redis import RedisError
from requests import RequestException

from renku.core.errors import (
    AuthenticationError,
    DatasetExistsError,
    DatasetImageError,
    DatasetNotFound,
    DockerfileUpdateError,
    GitCommandError,
    GitError,
    InvalidTemplateError,
    MetadataCorruptError,
    MigrationError,
    MigrationRequired,
    MinimumVersionError,
    ParameterError,
    ProjectNotFound,
    RenkuException,
    TemplateMissingReferenceError,
    TemplateUpdateError,
    UninitializedProject,
    WorkflowNotFoundError,
)
from renku.ui.service.errors import (
    IntermittentAuthenticationError,
    IntermittentDatasetExistsError,
    IntermittentFileNotExistsError,
    IntermittentProjectIdError,
    IntermittentProjectTemplateUnavailable,
    IntermittentRedisError,
    IntermittentSettingExistsError,
    IntermittentTimeoutError,
    IntermittentWorkflowNotFound,
    ProgramGitError,
    ProgramGraphCorruptError,
    ProgramInternalError,
    ProgramInvalidGenericFieldsError,
    ProgramProjectCorruptError,
    ProgramProjectCreationError,
    ProgramRenkuError,
    ProgramRepoUnknownError,
    ProgramUpdateProjectError,
    ServiceError,
    UserDatasetsMultipleImagesError,
    UserDatasetsNotFoundError,
    UserDatasetsUnlinkError,
    UserDatasetsUnreachableImageError,
    UserInvalidGenericFieldsError,
    UserMissingFieldError,
    UserNewerRenkuProjectError,
    UserNonRenkuProjectError,
    UserOutdatedProjectError,
    UserProjectCreationError,
    UserProjectMetadataCorruptError,
    UserProjectTemplateReferenceError,
    UserRepoBranchInvalidError,
    UserRepoNoAccessError,
    UserRepoUrlInvalidError,
    UserTemplateInvalidError,
)
from renku.ui.service.utils.squash import squash
from renku.ui.service.views import error_response


def handle_redis_except(f):
    """Wrapper which handles Redis exceptions."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (RedisError, OSError) as e:
            raise IntermittentRedisError(e)

    return decorated_function


def handle_validation_except(f):
    """Wrapper which handles marshmallow `ValidationError`."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            if isinstance(e.messages, dict):
                items = squash(e.messages).items()
                reasons = []
                for key, value in items:
                    if key == "project_id":
                        raise IntermittentProjectIdError(e)
                    reasons.append(f"'{key}': {', '.join(value)}")
            else:
                reasons = e.messages

            error_message = f"{'; '.join(reasons)}"
            if "Invalid `git_url`" in error_message:
                raise UserRepoUrlInvalidError(e, error_message)
            if "Unknown field" in error_message:
                raise ProgramInvalidGenericFieldsError(e, error_message)

            raise UserInvalidGenericFieldsError(e, error_message)

    return decorated_function


def handle_jwt_except(f):
    """Wrapper which handles invalid JWT."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (AuthenticationError, ExpiredSignatureError, ImmatureSignatureError, InvalidIssuedAtError) as e:
            raise IntermittentAuthenticationError(e)

    return decorated_function


def handle_renku_except(f):
    """Wrapper which handles `RenkuException`."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except MetadataCorruptError as e:
            raise UserProjectMetadataCorruptError(e, str(e))
        except MigrationRequired as e:
            raise UserOutdatedProjectError(e)
        except UninitializedProject as e:
            raise UserNonRenkuProjectError(e)
        except MinimumVersionError as e:
            raise UserNewerRenkuProjectError(e, minimum_version=e.minimum_version, current_version=e.current_version)
        except RenkuException as e:
            raise ProgramRenkuError(e)

    return decorated_function


def handle_git_except(f):
    """Wrapper which handles `RenkuException`."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:
            error_message = e.stderr.lower() if e.stderr else ""
            error_message_safe = format(" ".join(error_message.strip().split("\n")))
            error_message_safe = re.sub("^(.+oauth2:)[^@]+(@.+)$", r"\1<token-hidden>\2", error_message_safe)
            if "access denied" in error_message:
                raise UserRepoNoAccessError(e, error_message_safe)
            elif (
                "is this a git repository?" in error_message
                or "not found" in error_message
                or "ailed to connect to" in error_message  # Sometimes the 'f' is capitalized, sometimes not
            ):
                raise UserRepoUrlInvalidError(e, error_message_safe)
            elif "connection timed out" in error_message:
                raise IntermittentTimeoutError(e)
            else:
                raise ProgramRepoUnknownError(e, error_message_safe)

    return decorated_function


def handle_base_except(f):
    """Wrapper which handles base exceptions."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)

        # NOTE: HTTPException are now handled in the entrypoint
        except ServiceError as e:
            return error_response(e)
        # NOTE: GitError here may not be necessary anymore
        except GitError as e:
            return error_response(ProgramGitError(e, cast(str, e.message) if hasattr(e, "message") else ""))
        except (Exception, BaseException, OSError) as e:
            if hasattr(e, "stderr") and e.stderr:
                error_message = " ".join(e.stderr.strip().split("\n"))
            else:
                error_message = str(e)

            return error_response(ProgramInternalError(e, error_message))

    return decorated_function


def handle_common_except(f):
    """Handle common exceptions."""

    @wraps(f)
    def dec(*args, **kwargs):
        """Decorated function."""

        @handle_base_except
        @handle_validation_except
        @handle_renku_except
        @handle_git_except
        @handle_jwt_except
        def _wrapped(*args_, **kwargs_):
            return f(*args_, **kwargs_)

        return _wrapped(*args, **kwargs)

    return dec


def handle_templates_read_errors(f):
    """Wrapper which handles reading templates errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except InvalidTemplateError as e:
            raise UserTemplateInvalidError(e)
        except GitError as e:
            error_message = str(e)
            if "Cannot clone repo from" in error_message:
                raise UserRepoUrlInvalidError(e)
            elif "Cannot checkout reference" in error_message:
                raise UserRepoBranchInvalidError(e)
            elif "Cannot checkout manifest file" in error_message:
                raise UserTemplateInvalidError(e)
            raise

    return decorated_function


@handle_templates_read_errors
def handle_templates_create_errors(f):
    """Wrapper which handles template creating projects errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""

        def get_schema_error_message(e):
            if isinstance(getattr(e, "messages", None), dict) and e.messages.get("_schema"):
                message = (
                    "; ".join(e.messages.get("_schema"))
                    if isinstance(e.messages.get("_schema"), list)
                    else str(e.messages.get("_schema"))
                )
                return message
            return None

        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            if getattr(e, "field_name", None) == "_schema":
                error_message = get_schema_error_message(e)
                if error_message:
                    raise UserProjectCreationError(e, error_message)
                else:
                    raise ProgramProjectCreationError(e, str(e))
            else:
                raise ProgramProjectCreationError(e)

    return decorated_function


def handle_project_write_errors(f):
    """Wrapper which handles writing project metadata errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except UserInvalidGenericFieldsError as e:
            if "Unknown field" in str(e):
                raise ProgramInvalidGenericFieldsError(e)
            raise

    return decorated_function


def handle_config_read_errors(f):
    """Wrapper which handles reading config errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except Exception as e:
            if str(e).startswith("Source contains parsing errors"):
                raise ProgramProjectCorruptError(e)
            raise

    return decorated_function


@handle_config_read_errors
def handle_config_write_errors(f):
    """Wrapper which handles setting config errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except ParameterError as e:
            if "Invalid parameter value" in str(e):
                parameter = None
                match = re.search('Key "(.*)" not found', str(e))
                if match is not None and len(match.groups()) > 0:
                    parameter = match.group(1)
                raise IntermittentSettingExistsError(e, parameter)
            raise

    return decorated_function


def handle_datasets_write_errors(f):
    """Wrapper which handles datasets write errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except DatasetImageError as e:
            error_message = str(e)
            if "Duplicate dataset image" in error_message:
                raise UserDatasetsMultipleImagesError(e)
            elif "Cannot download image with url" in error_message:
                raise UserDatasetsUnreachableImageError(e)
            raise
        except ValidationError as e:
            items = squash(e.messages).items()
            for key, value in items:
                if "".join(value) == "Field may not be null.":
                    raise UserMissingFieldError(e, key)
            raise
        except DatasetNotFound as e:
            raise UserDatasetsNotFoundError(e)
        except DatasetExistsError as e:
            raise IntermittentDatasetExistsError(e)
        except RenkuException as e:
            if str(e).startswith("invalid file reference"):
                raise IntermittentFileNotExistsError(e)
            raise

    return decorated_function


def handle_workflow_errors(f):
    """Wrapper which handles workflow errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except WorkflowNotFoundError as e:
            raise IntermittentWorkflowNotFound(e, name_or_id=e.name_or_id)
        except RenkuException:
            raise

    return decorated_function


def handle_datasets_unlink_errors(f):
    """Wrapper which handles datasets unlink errors."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except RenkuException as e:
            if str(e).startswith("Invalid parameter") and "No records found" in str(e):
                raise UserDatasetsUnlinkError(e)
            raise
        except ValueError as e:
            if "one of the filters must be specified" in str(e):
                raise UserDatasetsUnlinkError(e)
            raise

    return decorated_function


def handle_migration_read_errors(f):
    """Wrapper which handles migrations read exceptions."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except TemplateMissingReferenceError as e:
            raise UserProjectTemplateReferenceError(e)
        except (InvalidTemplateError, TemplateUpdateError) as e:
            raise IntermittentProjectTemplateUnavailable(e)
        except ProjectNotFound as e:
            raise UserRepoUrlInvalidError(e)

    return decorated_function


@handle_migration_read_errors
def handle_migration_write_errors(f):
    """Wrapper which handles migrations write exceptions."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (TemplateUpdateError, DockerfileUpdateError, MigrationError) as e:
            raise ProgramUpdateProjectError(e)

    return decorated_function


def handle_graph_errors(f):
    """Wrapper which handles graph exceptions."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        # TODO: handle user and program errors separately
        except (MemoryError, RenkuException, RequestException) as e:
            raise ProgramGraphCorruptError(e)

    return decorated_function


def pretty_print_error(error: Exception):
    """Use error handlers to pretty print an exception."""

    @handle_common_except
    @handle_migration_read_errors
    def _fake_error_source():
        raise error

    response = _fake_error_source()
    return response.json["error"]



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Support JSON-LD context in models."""

from pathlib import Path
from typing import Union

import yaml

try:
    from yaml import CDumper as Dumper
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import Dumper, SafeLoader  # type: ignore


# Shamelessly copy/pasting from SO:
# https://stackoverflow.com/questions/34667108/ignore-dates-and-times-while-parsing-yaml
# This is needed to allow us to load from yaml and use json down the line.
class NoDatesSafeLoader(SafeLoader):
    """Used to safely load basic python objects but ignore datetime strings."""

    @classmethod
    def remove_implicit_resolver(cls, tag_to_remove):
        """
        Remove implicit resolvers for a particular tag.

        Takes care not to modify resolvers in super classes.

        We want to load datetimes as strings, not dates, because we
        go on to serialize as json which doesn't have the advanced types
        of yaml, and leads to incompatibilities down the track.
        """
        if "yaml_implicit_resolvers" not in cls.__dict__:
            cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

        for first_letter, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[first_letter] = [
                (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
            ]


NoDatesSafeLoader.remove_implicit_resolver("tag:yaml.org,2002:timestamp")


def read_yaml(path: Union[Path, str]):
    """Load YAML file and return its content as a dict."""
    with Path(path).open(mode="r") as fp:
        return load_yaml(fp)


def write_yaml(path, data):
    """Store data to a YAML file."""

    Dumper.ignore_aliases = lambda _, data: True  # type: ignore

    with Path(path).open("w") as fp:
        yaml.dump(data, fp, default_flow_style=False, Dumper=Dumper)


def dumps_yaml(data) -> str:
    """Convert YAML data to a YAML string."""
    return yaml.dump(data)


def load_yaml(data):
    """Load YAML data and return its content as a dict."""
    return yaml.load(data, Loader=NoDatesSafeLoader) or {}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Workflow Models.

Input and Output classes can be used to define inputs and outputs of a script
within the same script. Paths defined with these classes are added to explicit
inputs and outputs in the workflow's metadata. For example, the following
mark a ``data/data.csv`` as an input with name ``my-input`` to the script:

.. code-block:: python

    from renku.api import Input

    with open(Input("my-input", "data/data.csv")) as input_data:
        for line in input_data:
            print(line)


Users can track parameters' values in a workflow by defining them using
``Parameter`` function.

.. code-block:: python

    from renku.api import Parameter

    nc = Parameter(name="n_components", value=10)

    print(nc.value)  # 10

Once a Parameter is tracked like this, it can be set normally in commands like
``renku workflow execute`` with the ``--set`` option to override the value.

"""

from os import PathLike, environ
from pathlib import Path
from typing import Any, List, Optional, Union

from renku.core import errors
from renku.core.plugin.provider import RENKU_ENV_PREFIX
from renku.core.workflow.plan_factory import (
    add_indirect_parameter,
    add_to_files_list,
    get_indirect_inputs_path,
    get_indirect_outputs_path,
)
from renku.domain_model.workflow import parameter as core_parameter
from renku.domain_model.workflow import plan as core_plan
from renku.ui.api.util import ensure_project_context


def _validate_name(name: str):
    if not name:
        raise errors.ParameterError("'name' must be set.")

    core_plan.Plan.validate_name(name)


class _PathBase(PathLike):
    """Base class of API input/output parameters."""

    @ensure_project_context
    def __init__(self, name: str, path: Union[str, Path], project=None, skip_addition: bool = False):
        env_value = environ.get(f"{RENKU_ENV_PREFIX}{name}", None)
        if env_value:
            self._path = Path(env_value)
        else:
            self._path = Path(path)

            if not skip_addition:
                indirect_list_path = self._get_indirect_list_path(project.path)
                add_to_files_list(indirect_list_path, name, self._path)

    @staticmethod
    def _get_indirect_list_path(project_path):
        raise NotImplementedError

    def __fspath__(self):
        """Abstract method of PathLike."""
        return str(self._path)

    @property
    def path(self) -> Path:
        """Return path of file."""
        return self._path


class Parameter:
    """API Parameter model."""

    @ensure_project_context
    def __init__(self, name: str, value: Union[Path, str, bool, int, float], project, skip_addition: bool = False):
        _validate_name(name)

        value = self._get_parameter_value(name=name, value=value)

        self.name: str = name
        self.value: Union[Path, str, bool, int, float] = value
        self.default_value: Union[Path, str, bool, int, float] = value
        self.description: Optional[str] = None
        self.position: Optional[int] = None
        self.prefix: Optional[str] = None

        if not skip_addition:
            add_indirect_parameter(project.path, name=name, value=value)

    @classmethod
    def from_parameter(cls, parameter: core_parameter.CommandParameter) -> "Parameter":
        """Create an instance from a core CommandParameterBase."""
        value = parameter.actual_value
        default_value = parameter.default_value
        value = value if value is not None else default_value

        self = cls(name=parameter.name, value=value, skip_addition=True)
        self.default_value = default_value if default_value is not None else value
        self.description = parameter.description
        self.position = parameter.position
        self.prefix = parameter.prefix

        return self

    @staticmethod
    def _get_parameter_value(name: str, value: Any) -> Union[str, bool, int, float]:
        """Get parameter's actual value from env variables.

        Args:
            name (str): The name of the parameter.
            value (Any): The value of the parameter.

        Returns:
            The supplied value or a value set on workflow execution.
        """
        env_value = environ.get(f"{RENKU_ENV_PREFIX}{name}", None)

        if env_value:
            if isinstance(value, str):
                value = env_value
            elif isinstance(value, bool):
                value = bool(env_value)
            elif isinstance(value, int):
                value = int(env_value)
            elif isinstance(value, float):
                value = float(env_value)
            else:
                raise errors.ParameterError(
                    f"Can't convert value '{env_value}' to type '{type(value)}' for parameter '{name}'. Only "
                    "str, bool, int and float are supported."
                )

        return value

    def __repr__(self):
        return (
            f"<Parameter '{self.name}': {self.value} (default: {self.default_value}, prefix: {self.prefix}, "
            f"position: {self.position})>"
        )


class Input(_PathBase, Parameter):
    """API Input model."""

    @staticmethod
    def _get_indirect_list_path(project_path):
        return get_indirect_inputs_path(project_path)

    def __init__(self, name: str, path: Union[str, Path], skip_addition: bool = False):
        _PathBase.__init__(self, name=name, path=path, skip_addition=skip_addition)
        Parameter.__init__(self, name=name, value=path, skip_addition=True)

        self.mapped_stream: Optional[str] = None

    @classmethod
    def from_parameter(cls, input: core_parameter.CommandInput) -> "Input":  # type: ignore
        """Create an instance from a CommandInput."""
        assert isinstance(input, core_parameter.CommandInput)

        self = cls(name=input.name, path=input.default_value, skip_addition=True)
        self.description = input.description
        self.prefix = input.prefix
        self.position = input.position
        self.mapped_stream = input.mapped_to.stream_type if input.mapped_to else None

        return self

    def __repr__(self):
        return f"<Input '{self.name}'={self.path}>"


class Output(_PathBase, Parameter):
    """API Output model."""

    @staticmethod
    def _get_indirect_list_path(project_path):
        return get_indirect_outputs_path(project_path)

    def __init__(self, name: str, path: Union[str, Path], skip_addition: bool = False):
        _PathBase.__init__(self, name=name, path=path, skip_addition=skip_addition)
        Parameter.__init__(self, name=name, value=path, skip_addition=True)

        self.mapped_stream: Optional[str] = None

    @classmethod
    def from_parameter(cls, output: core_parameter.CommandOutput) -> "Output":  # type: ignore
        """Create an instance from a CommandOutput."""
        assert isinstance(output, core_parameter.CommandOutput)

        self = cls(name=output.name, path=output.default_value, skip_addition=True)
        self.description = output.description
        self.prefix = output.prefix
        self.position = output.position
        self.mapped_stream = output.mapped_to.stream_type if output.mapped_to else None

        return self

    def __repr__(self):
        return f"<Output '{self.name}'={self.path}>"


class Link:
    """Parameter Link API model."""

    def __init__(self, source: Parameter, sinks: List[Parameter]):
        self.source: Parameter = source
        self.sinks: List[Parameter] = sinks

    @classmethod
    def from_link(cls, link: core_parameter.ParameterLink) -> "Link":
        """Create an instance from a ParameterLink."""
        return cls(source=convert_parameter(link.source), sinks=[convert_parameter(p) for p in link.sinks])


class Mapping(Parameter):
    """Parameter Mapping API model."""

    def __init__(
        self,
        name: str,
        value: Union[Path, str, bool, int, float],
        default_value: Optional[Union[Path, str, bool, int, float]] = None,
        description: Optional[str] = None,
        parameters: Optional[List[Parameter]] = None,
    ):
        super().__init__(name=name, value=value)
        self.default_value: Union[Path, str, bool, int, float] = default_value if default_value is not None else value
        self.description: Optional[str] = description
        self.parameters: List[Parameter] = parameters or []

    @classmethod
    def from_parameter(cls, mapping: core_parameter.CommandParameterBase) -> "Mapping":
        """Create an instance from a ParameterMapping."""
        assert isinstance(mapping, core_parameter.ParameterMapping)

        return cls(
            name=mapping.name,
            value=mapping.actual_value,
            default_value=mapping.default_value,
            description=mapping.description,
            parameters=[convert_parameter(p) for p in mapping.mapped_parameters],
        )


def convert_parameter(
    parameter: Union[core_parameter.CommandParameterBase],
) -> Union[Input, Output, Parameter, Mapping]:
    """Convert a core CommandParameterBase subclass to its equivalent API class."""
    if isinstance(parameter, core_parameter.CommandInput):
        return Input.from_parameter(parameter)
    elif isinstance(parameter, core_parameter.CommandOutput):
        return Output.from_parameter(parameter)
    elif isinstance(parameter, core_parameter.CommandParameter):
        return Parameter.from_parameter(parameter)
    elif isinstance(parameter, core_parameter.ParameterMapping):
        return Mapping.from_parameter(parameter)

    raise errors.ParameterError(f"Invalid parameter type: '{type(parameter)}'")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder."""

import contextlib
import functools
import threading
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, Union

import inject

from renku.core import errors
from renku.core.util.communication import CommunicationCallback
from renku.core.util.git import get_git_path
from renku.domain_model.project_context import project_context

_LOCAL = threading.local()


def check_finalized(f):
    """Decorator to prevent modification of finalized builders.

    Args:
        f: Decorated function.

    Returns:
        Wrapped function.
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """Decorator to prevent modification of finalized builders."""
        if not args or not isinstance(args[0], Command):
            raise errors.ParameterError("Command hooks need to be `Command` object methods.")

        if args[0].finalized:
            raise errors.CommandFinalizedError("Cannot modify a finalized `Command`.")

        return f(*args, **kwargs)

    return wrapper


def _patched_get_injector_or_die() -> inject.Injector:
    """Patched version of get_injector_or_die with thread local injectors.

    Allows deferred definition of an injector per thread.
    """
    injector = getattr(_LOCAL, "injector", None)
    if not injector:
        raise inject.InjectorException("No injector is configured")

    return injector


def _patched_configure(config: Optional[inject.BinderCallable] = None, bind_in_runtime: bool = True) -> inject.Injector:
    """Create an injector with a callable config or raise an exception when already configured.

    Args:
        config(Optional[inject.BinderCallable], optional): Injection binding config (Default value = None).
        bind_in_runtime(bool, optional): Whether to allow binding at runtime (Default value = True).

    Returns:
        Injector: Thread-safe injector with bindings applied.
    """

    if getattr(_LOCAL, "injector", None):
        raise inject.InjectorException("Injector is already configured")

    _LOCAL.injector = inject.Injector(config, bind_in_runtime=bind_in_runtime)

    return _LOCAL.injector


inject.configure = _patched_configure
inject.get_injector_or_die = _patched_get_injector_or_die


def remove_injector():
    """Remove a thread-local injector."""
    if getattr(_LOCAL, "injector", None):
        del _LOCAL.injector


@contextlib.contextmanager
def replace_injection(bindings: Dict, constructor_bindings=None):
    """Temporarily inject various test objects.

    Args:
        bindings: New normal injection bindings to apply.
        constructor_bindings: New constructor bindings to apply (Default value = None).
    """
    constructor_bindings = constructor_bindings or {}

    def bind(binder):
        for key, value in bindings.items():
            binder.bind(key, value)
        for key, value in constructor_bindings.items():
            binder.bind_to_constructor(key, value)

    old_injector = getattr(_LOCAL, "injector", None)
    try:
        if old_injector:
            remove_injector()
        inject.configure(bind, bind_in_runtime=False)

        yield
    finally:
        remove_injector()

        if old_injector:
            _LOCAL.injector = old_injector


class Command:
    """Base renku command builder."""

    HOOK_ORDER = 1

    def __init__(self) -> None:
        """__init__ of Command."""
        self.injection_pre_hooks: Dict[int, List[Callable]] = defaultdict(list)
        self.pre_hooks: Dict[int, List[Callable]] = defaultdict(list)
        self.post_hooks: Dict[int, List[Callable]] = defaultdict(list)
        self._operation: Optional[Callable] = None
        self._finalized: bool = False
        self._track_std_streams: bool = False
        self._working_directory: Optional[str] = None
        self._context_added: bool = False

    def __getattr__(self, name: str) -> Any:
        """Bubble up attributes of wrapped builders."""
        if "_builder" in self.__dict__:
            return getattr(self._builder, name)

        raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set attributes of wrapped builders."""
        if hasattr(self, "_builder") and self.__class__ is not self._builder.__class__:
            self._builder.__setattr__(name, value)

        object.__setattr__(self, name, value)

    def _injection_pre_hook(self, builder: "Command", context: dict, *args, **kwargs) -> None:
        """Setup dependency injections.

        Args:
            builder("Command"): Current ``CommandBuilder``.
            context(dict): Current context dictionary.
        """
        if not project_context.has_context():
            path = get_git_path(self._working_directory or ".")
            project_context.push_path(path)
            self._context_added = True

        context["bindings"] = {}
        context["constructor_bindings"] = {}

    def _pre_hook(self, builder: "Command", context: dict, *args, **kwargs) -> None:
        """Setup project.

        Args:
            builder("Command"): Current ``CommandBuilder``.
            context(dict): Current context dictionary.
        """

        stack = contextlib.ExitStack()
        context["stack"] = stack

    def _post_hook(self, builder: "Command", context: dict, result: "CommandResult", *args, **kwargs) -> None:
        """Post-hook method.

        Args:
            builder("Command"): Current ``CommandBuilder``.
            context(dict): Current context dictionary.
            result("CommandResult"): Result of command execution.
        """
        remove_injector()

        if self._context_added:
            project_context.pop_context()

        if result.error:
            raise result.error

    def execute(self, *args, **kwargs) -> "CommandResult":
        """Execute the wrapped operation.

        First executes `pre_hooks` in ascending `order`, passing a read/write context between them.
        It then calls the wrapped `operation`. The result of the operation then gets pass to all the `post_hooks`,
        but in descending `order`. It then returns the result or error if there was one.

        Returns:
            CommandResult: Result of execution of command.
        """
        if not self.finalized:
            raise errors.CommandNotFinalizedError("Call `build()` before executing a command")

        context: Dict[str, Any] = {}
        if any(self.injection_pre_hooks):
            order = sorted(self.injection_pre_hooks.keys())

            for o in order:
                for hook in self.injection_pre_hooks[o]:
                    hook(self, context, *args, **kwargs)

        def _bind(binder):
            for key, value in context["bindings"].items():
                binder.bind(key, value)
            for key, value in context["constructor_bindings"].items():
                binder.bind_to_constructor(key, value)

            return binder

        inject.configure(_bind, bind_in_runtime=False)

        if any(self.pre_hooks):
            order = sorted(self.pre_hooks.keys())

            for o in order:
                for hook in self.pre_hooks[o]:
                    try:
                        hook(self, context, *args, **kwargs)
                    except (Exception, BaseException):
                        # don't leak injections from failed hook
                        remove_injector()
                        raise

        output = None
        error = None

        try:
            with context["stack"]:
                output = self._operation(*args, **kwargs)  # type: ignore
        except errors.RenkuException as e:
            error = e
        except (Exception, BaseException):
            remove_injector()
            raise

        result = CommandResult(output, error, CommandResult.FAILURE if error else CommandResult.SUCCESS)

        if any(self.post_hooks):
            order = sorted(self.post_hooks.keys(), reverse=True)

            for o in order:
                for hook in self.post_hooks[o]:
                    hook(self, context, result, *args, **kwargs)

        return result

    @property
    def finalized(self) -> bool:
        """Whether this builder is still being constructed or has been finalized."""
        if hasattr(self, "_builder"):
            return self._builder.finalized
        return self._finalized

    def any_builder_is_instance_of(self, cls: Type) -> bool:
        """Check if any 'chained' command builder is an instance of a specific command builder class."""
        if isinstance(self, cls):
            return True
        elif "_builder" in self.__dict__:
            return self._builder.any_builder_is_instance_of(cls)
        else:
            return False

    @property
    def will_write_to_database(self) -> bool:
        """Will running the command write anything to the metadata store."""
        try:
            return self._write
        except AttributeError:
            return False

    @check_finalized
    def add_injection_pre_hook(self, order: int, hook: Callable):
        """Add a pre-execution hook for dependency injection.

        Args:
            order(int): Determines the order of executed hooks, lower numbers get executed first.
            hook(Callable): The hook to add.
        """
        if hasattr(self, "_builder"):
            self._builder.add_injection_pre_hook(order, hook)
        else:
            self.injection_pre_hooks[order].append(hook)

    @check_finalized
    def add_pre_hook(self, order: int, hook: Callable):
        """Add a pre-execution hook.

        Args:
            order(int): Determines the order of executed hooks, lower numbers get executed first.
            hook(Callable): The hook to add.
        """
        if hasattr(self, "_builder"):
            self._builder.add_pre_hook(order, hook)
        else:
            self.pre_hooks[order].append(hook)

    @check_finalized
    def add_post_hook(self, order: int, hook: Callable):
        """Add a post-execution hook.

        Args:
            order(int): Determines the order of executed hooks, higher numbers get executed first.
            hook(Callable): The hook to add.
        """
        if hasattr(self, "_builder"):
            self._builder.add_post_hook(order, hook)
        else:
            self.post_hooks[order].append(hook)

    @check_finalized
    def build(self) -> "Command":
        """Build (finalize) the command.

        Returns:
            Command: Finalized command that cannot be modified.
        """
        if not self._operation:
            raise errors.ConfigurationError("`Command` needs to have a wrapped `command` set")
        self.add_injection_pre_hook(self.HOOK_ORDER, self._injection_pre_hook)
        self.add_pre_hook(self.HOOK_ORDER, self._pre_hook)
        self.add_post_hook(self.HOOK_ORDER, self._post_hook)

        self._finalized = True

        return self

    @check_finalized
    def command(self, operation: Callable):
        """Set the wrapped command.

        Args:
            operation(Callable): The function to wrap in the command builder.

        Returns:
            Command: This command.
        """

        self._operation = operation

        return self

    @check_finalized
    def working_directory(self, directory: str) -> "Command":
        """Set the working directory for the command.

        WARNING: Should not be used in the core service.

        Args:
            directory(str): The working directory to work in.

        Returns:
            Command: This command.
        """
        self._working_directory = directory

        return self

    @check_finalized
    def track_std_streams(self) -> "Command":
        """Whether to track STD streams or not.

        Returns:
            Command: This command.
        """
        self._track_std_streams = True

        return self

    @check_finalized
    def with_git_isolation(self) -> "Command":
        """Whether to run in git isolation or not."""
        from renku.command.command_builder.repo import Isolation

        return Isolation(self)

    @check_finalized
    def with_commit(
        self,
        message: Optional[str] = None,
        commit_if_empty: bool = False,
        raise_if_empty: bool = False,
        commit_only: Optional[Union[str, List[Union[str, Path]]]] = None,
        skip_staging: bool = False,
        skip_dirty_checks: bool = False,
    ) -> "Command":
        """Create a commit.

        Args:
            message(str, optional): The commit message. Auto-generated if left empty (Default value = None).
            commit_if_empty(bool, optional): Whether to commit if there are no modified files (Default value = False).
            raise_if_empty(bool, optional): Whether to raise an exception if there are no modified files
                (Default value = False).
            commit_only(bool, optional): Only commit the supplied paths (Default value = None).
            skip_staging(bool): Don't commit staged files.
            skip_dirty_checks(bool): Don't check if paths are dirty or staged.
        """
        from renku.command.command_builder.repo import Commit

        return Commit(
            self,
            message=message,
            commit_if_empty=commit_if_empty,
            raise_if_empty=raise_if_empty,
            commit_only=commit_only,
            skip_staging=skip_staging,
            skip_dirty_checks=skip_dirty_checks,
        )

    @check_finalized
    def lock_project(self) -> "Command":
        """Acquire a lock for the whole project."""
        from renku.command.command_builder.lock import ProjectLock

        return ProjectLock(self)

    @check_finalized
    def lock_dataset(self) -> "Command":
        """Acquire a lock for a dataset."""
        from renku.command.command_builder.lock import DatasetLock

        return DatasetLock(self)

    @check_finalized
    def require_migration(self) -> "Command":
        """Check if a migration is needed."""
        from renku.command.command_builder.migration import RequireMigration

        return RequireMigration(self)

    @check_finalized
    def require_clean(self) -> "Command":
        """Check that the repository is clean."""
        from renku.command.command_builder.repo import RequireClean

        return RequireClean(self)

    @check_finalized
    def with_communicator(self, communicator: CommunicationCallback) -> "Command":
        """Create a communicator.

        Args:
            communicator(CommunicationCallback): Communicator to use for writing to user.
        """
        from renku.command.command_builder.communication import Communicator

        return Communicator(self, communicator)

    @check_finalized
    def with_database(self, write: bool = False, path: Optional[str] = None, create: bool = False) -> "Command":
        """Provide an object database connection.

        Args:
            write(bool, optional): Whether or not to persist changes to the database (Default value = False).
            path(str, optional): Location of the database (Default value = None).
            create(bool, optional): Whether the database should be created if it doesn't exist (Default value = False).
        """
        from renku.command.command_builder.database import DatabaseCommand

        return DatabaseCommand(self, write, path, create)


class CommandResult:
    """The result of a command.

    The return value of the command is set as `.output`, if there was an error, it is set as `.error`, and
    the status of the command is set to either `CommandResult.SUCCESS` or CommandResult.FAILURE`.
    """

    SUCCESS = 0

    FAILURE = 1

    def __init__(self, output, error, status) -> None:
        """__init__ of CommandResult."""
        self.output = output
        self.error = error
        self.status = status



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent workflow file run templates."""

from itertools import chain
from typing import List, Union

import marshmallow

from renku.command.schema.calamus import fields, prov, renku, schema
from renku.command.schema.composite_plan import CompositePlanSchema
from renku.command.schema.plan import PlanSchema
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan, WorkflowFilePlan


class WorkflowFilePlanSchema(PlanSchema):
    """WorkflowFilePlan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork, renku.Plan, renku.WorkflowFilePlan]
        model = WorkflowFilePlan
        unknown = marshmallow.EXCLUDE

    @marshmallow.pre_dump(pass_many=True)
    def fix_ids(self, objs: Union[WorkflowFilePlan, List[WorkflowFilePlan]], many, **kwargs):
        """Renku up to 2.4.1 had a bug that created wrong ids for workflow file entities, this fixes those on export."""

        def _replace_id(obj):
            obj.unfreeze()
            obj.id = obj.id.replace("//plans/", "/")

            for child in chain(obj.inputs, obj.outputs, obj.parameters):
                child.id = child.id.replace("//plans/", "/")
            obj.freeze()

        if many:
            for obj in objs:
                _replace_id(obj)
            return objs

        _replace_id(objs)
        return objs


class WorkflowFileCompositePlanSchema(CompositePlanSchema):
    """Plan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork, renku.CompositePlan, renku.WorkflowFileCompositePlan]
        model = WorkflowFileCompositePlan
        unknown = marshmallow.EXCLUDE

    path = fields.String(prov.atLocation)
    plans = fields.Nested(renku.hasSubprocess, WorkflowFilePlanSchema, many=True)

    @marshmallow.pre_dump(pass_many=True)
    def fix_ids(self, objs, many, **kwargs):
        """Renku up to 2.4.1 had a bug that created wrong ids for workflow file entities, this fixes those on export."""

        def _replace_id(obj):
            obj.unfreeze()
            obj.id = obj.id.replace("//plans/", "/")
            obj.freeze()

        if many:
            for obj in objs:
                _replace_id(obj)
            return objs

        _replace_id(objs)
        return objs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git utilities."""

import os
import re
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import attr

from renku.core import errors
from renku.core.util.os import is_ascii, normalize_to_ascii
from renku.ui.service.utils import normalize_git_url

_RE_SCHEME = r"(?P<scheme>(git\+)?(https?|git|ssh|rsync))\://"

_RE_USERNAME = r"(?:(?P<username>.+)@)?"

_RE_USERNAME_PASSWORD = r"(?:(?P<username>[^:]+)(:(?P<password>[^@]+))?@)?"

# RFC 1123 compliant hostname regex
_RE_HOSTNAME = (
    r"(?P<hostname>"
    r"([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}"
    r"[a-zA-Z0-9])(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}"
    r"[a-zA-Z0-9]))*)"
)

_RE_PORT = r":(?P<port>\d+)"

_RE_PATHNAME = r"(?P<path>((\~[\w\-\.]+)/)*?(((?P<owner>[\w\-\./]+?)/)?(?P<name>[\w\-\.]+)(\.git)?)?)/*"

_RE_PATHNAME_WITH_GITLAB = (
    r"(?P<path>((((gitlab/){0,1}|([\w\-\~\.]+/)*?)(?P<owner>([\w\-\.]+/)*[\w\-\.]+)/)?"
    r"(?P<name>[\w\-\.]+)(\.git)?)?)/*"
)

_RE_UNIXPATH = r"(file\://)?(?P<path>\/$|((?=\/)|\.|\.\.)(\/(?=[^/\0])[^/\0]+)*\/?)"


def _build(*parts):
    """Assemble the regex."""
    return re.compile(r"^" + r"".join(parts) + r"$")


#: Define possible repository URLs.
_REPOSITORY_URLS = [
    # https://user:pass@example.com/owner/repo.git
    _build(_RE_SCHEME, _RE_USERNAME_PASSWORD, _RE_HOSTNAME, r"/", _RE_PATHNAME_WITH_GITLAB),
    # https://user:pass@example.com:gitlab/owner/repo.git
    _build(_RE_SCHEME, _RE_USERNAME_PASSWORD, _RE_HOSTNAME, _RE_PORT, r"/", _RE_PATHNAME_WITH_GITLAB),
    # git@example.com:owner/repo.git
    _build(_RE_USERNAME, _RE_HOSTNAME, r":", _RE_PATHNAME_WITH_GITLAB),
    # /path/to/repo
    _build(_RE_UNIXPATH),
]


@attr.s()
class GitURL:
    """Parser for common Git URLs."""

    # Initial value
    href = attr.ib()
    path = attr.ib(default=None)
    scheme = attr.ib(default="ssh")
    hostname = attr.ib(default="localhost")
    username = attr.ib(default=None)
    password = attr.ib(default=None)
    port = attr.ib(default=None)
    owner = attr.ib(default=None)
    name: Optional[str] = attr.ib(default=None, converter=normalize_git_url)
    slug = attr.ib(default=None)
    _regex = attr.ib(default=None, eq=False, order=False)

    def __attrs_post_init__(self):
        """Derive basic information."""
        if not self.name and self.path:
            self.name = normalize_git_url(Path(self.path).name)

        self.slug = normalize_to_ascii(self.name)

    @classmethod
    def parse(cls, href) -> "GitURL":
        """Derive URI components."""
        if not is_ascii(href):
            raise UnicodeError(f"`{href}` is not a valid Git remote")

        url_regexes = _REPOSITORY_URLS

        gitlab_url = os.environ.get("GITLAB_BASE_URL", None)

        if gitlab_url:
            # NOTE: use known gitlab url to simplify regex to make detection more robust
            parsed_gitlab_url = urlparse(gitlab_url)
            gitlab_re = _build(
                _RE_SCHEME,
                _RE_USERNAME_PASSWORD,
                r"(?P<hostname>" + re.escape(parsed_gitlab_url.hostname) + ")",  # type: ignore
                r":(?P<port>" + str(parsed_gitlab_url.port) + ")" if parsed_gitlab_url.port else "",
                r"/",
                re.escape(parsed_gitlab_url.path) + r"/" if parsed_gitlab_url.path else "",
                _RE_PATHNAME,
            )
            url_regexes = [gitlab_re] + url_regexes

        for regex in url_regexes:
            matches = re.search(regex, href)
            if matches:
                return cls(href=href, regex=regex, **matches.groupdict())
        else:
            raise errors.InvalidGitURL(f"`{href}` is not a valid Git URL")

    @property
    def instance_url(self):
        """Get the url of the git instance."""
        url = urlparse(self.href)

        path = self.path.split(self.owner, 1)[0]
        url = url._replace(path=path)

        return url.geturl()

    @property
    def image(self):
        """Return image name."""
        img = self.hostname
        if self.owner:
            img += "/" + self.owner
        if self.name:
            img += "/" + self.name
        return img



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku environment related commands."""

import click


@click.command()
@click.option("shell_completion", "--shell-completion", is_flag=True, help="Print shell completion command")
@click.pass_context
def env(ctx, shell_completion):
    """Renku environment commands."""

    if shell_completion:
        import shellingham

        from renku.core.errors import UsageError

        shell = shellingham.detect_shell()[0]
        if shell == "bash":
            click.echo("_RENKU_COMPLETE=bash_source renku")
        elif shell == "zsh":
            click.echo("_RENKU_COMPLETE=zsh_source renku")
        elif shell == "fish":
            click.echo("env _RENKU_COMPLETE=fish_source renku")
        else:
            raise UsageError(f"The currently used shell '{shell}' is not supported for shell completion.")
    else:
        click.echo(ctx.get_help())



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder for communication."""

from renku.command.command_builder.command import Command, CommandResult, check_finalized
from renku.core.util import communication


class Communicator(Command):
    """Hook for logging and interaction with user."""

    HOOK_ORDER = 2

    def __init__(self, builder: Command, communicator: communication.CommunicationCallback) -> None:
        """__init__ of Communicator.

        Args:
            communicator: Instance of CommunicationCallback.
        """
        self._builder = builder
        self._communicator = communicator

    def _pre_hook(self, builder: Command, context: dict, *args, **kwargs) -> None:
        communication.subscribe(self._communicator)

    def _post_hook(self, builder: Command, context: dict, result: CommandResult, *args, **kwargs) -> None:
        communication.unsubscribe(self._communicator)

    @check_finalized
    def build(self) -> Command:
        """Build the command.

        Returns:
            Command: Finalized version of this command.
        """
        self._builder.add_pre_hook(self.HOOK_ORDER, self._pre_hook)
        self._builder.add_post_hook(self.HOOK_ORDER, self._post_hook)

        return self._builder.build()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project business logic."""

import os
import shutil
from typing import Dict, List, Optional, Union, cast

from pydantic import ConfigDict, validate_call

from renku.command.command_builder import inject
from renku.command.view_model.project import ProjectViewModel
from renku.core import errors
from renku.core.image import ImageObjectRequest
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util.metadata import construct_creator
from renku.core.util.os import get_relative_path
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.domain_model.dataset import ImageObjectRequestJson
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.agent import Person


@inject.autoparams()
@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def edit_project(
    description: Optional[Union[str, NoValueType]],
    creator: Union[Dict, str, NoValueType],
    keywords: Optional[Union[List[str], NoValueType]],
    custom_metadata: Optional[Union[Dict, List[Dict], NoValueType]],
    custom_metadata_source: Optional[Union[str, NoValueType]],
    image_request: Optional[Union[ImageObjectRequest, NoValueType]],
    project_gateway: IProjectGateway,
):
    """Edit dataset metadata.

    Args:
        description(Union[Optional[str], NoValueType]): New description.
        creator(Union[Dict, str, NoValueType]): New creators.
        keywords(Union[Optional[List[str]]): New keywords.
        custom_metadata(Union[Optional[Dict, List[Dict]]): Custom JSON-LD metadata.
        custom_metadata_source(Optional[str]): Custom metadata source.
        project_gateway(IProjectGateway): Injected project gateway.
        image_request(Optional[ImageObjectRequest]): Project's image.

    Returns:
        Tuple of fields that were updated and dictionary of warnings.
    """
    possible_updates = {
        "creator": creator,
        "description": description,
        "keywords": keywords,
        "custom_metadata": custom_metadata,
        "image": (
            image_request
            if image_request is NO_VALUE or image_request is None
            else ImageObjectRequestJson().dump(image_request)
        ),
    }

    no_email_warnings: Optional[Union[Dict, str]] = None
    parsed_creator: Optional[Union[NoValueType, Person]] = NO_VALUE

    if creator is not NO_VALUE:
        parsed_creator, no_email_warnings = construct_creator(cast(Union[Dict, str], creator), ignore_email=True)

    if image_request is None:
        delete_project_image()
    elif image_request is not NO_VALUE:
        set_project_image(image_request=image_request)  # type: ignore

    updated = {k: v for k, v in possible_updates.items() if v is not NO_VALUE}

    if updated:
        project = project_gateway.get_project()
        # NOTE: No need to pass ``image`` here since we already copied/deleted the file and updated the project
        project.update_metadata(
            creator=parsed_creator,
            description=description,
            keywords=keywords,
            custom_metadata=custom_metadata,
            custom_metadata_source=custom_metadata_source,
        )
        project_gateway.update_project(project)

    return updated, no_email_warnings


def show_project() -> ProjectViewModel:
    """Show project metadata.

    Returns:
        Project view model.
    """
    return ProjectViewModel.from_project(project_context.project)


def set_project_image(image_request: Optional[ImageObjectRequest]) -> None:
    """Download and set a project's images.

    Args:
        image_request(Optional[ImageObjectRequest]): The image to set.
    """
    if image_request is None:
        return

    # NOTE: Projects can have maximum one image
    image_request.position = 0

    image_object = image_request.to_image_object(owner_id=project_context.project.id)

    project_image = project_context.project_image_pathname

    # NOTE: Do nothing if the new path is the same as the old one
    if project_image.resolve() != image_object.content_url:
        # NOTE: Always delete the old image in case the image wasn't mirrored in the project
        delete_project_image()

        if not image_object.is_remote:
            project_image.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(image_object.content_url, project_context.project_image_pathname)

            image_object.content_url = get_relative_path(project_image, base=project_context.path)  # type: ignore

    project_context.project.image = image_object


def delete_project_image() -> None:
    """Delete project image in a project."""
    try:
        os.remove(project_context.project_image_pathname)
    except FileNotFoundError:
        pass
    except OSError as e:
        raise errors.ImageError(f"Cannot delete project image '{project_context.project_image_pathname}': {e}") from e
    else:
        project_context.project.image = None



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git repository management."""

import os
import sys
import tempfile
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Tuple, Type

from renku.core import errors
from renku.core.storage import checkout_paths_from_storage
from renku.core.util.contexts import Isolation
from renku.core.util.git import get_dirty_paths
from renku.core.util.os import get_absolute_path
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Repository


@contextmanager
def with_worktree(path=None, branch_name=None, commit=None, merge_args=("--ff-only",)):
    """Create new worktree."""
    from renku.infrastructure.repository import NULL_TREE

    delete = branch_name is None
    new_branch = commit is not NULL_TREE

    _, isolation, path, branch_name = prepare_worktree(path=path, branch_name=branch_name, commit=commit)
    try:
        yield
    except (Exception, BaseException) as e:
        finalize_worktree(
            isolation=isolation,
            path=path,
            branch_name=branch_name,
            delete=delete,
            new_branch=new_branch,
            merge_args=merge_args,
            exception=e,
        )
        raise
    else:
        finalize_worktree(
            isolation=isolation,
            path=path,
            branch_name=branch_name,
            delete=delete,
            new_branch=new_branch,
            merge_args=merge_args,
        )


def prepare_worktree(path=None, branch_name=None, commit=None) -> Tuple[Repository, Isolation, Path, str]:
    """Set up a Git worktree to provide isolation."""
    from renku.infrastructure.repository import NULL_TREE

    original_repository = project_context.repository
    original_path = project_context.path

    path = path or tempfile.mkdtemp()
    branch_name = branch_name or f"renku/run/isolation/{uuid.uuid4().hex}"

    project_context.push_path(path)
    path = project_context.path

    # TODO sys.argv
    if commit is NULL_TREE:
        original_repository.create_worktree(path=path, reference="HEAD", detach=True)
        repository = project_context.repository
        repository.run_git_command("checkout", "--orphan", branch_name)
        repository.remove("*", recursive=True, force=True)
    else:
        revision = None
        if commit:
            revision = commit.hexsha
        original_repository.create_worktree(path=path, branch=branch_name, reference=revision)
        repository = project_context.repository

    repository.get_configuration = original_repository.get_configuration  # type: ignore

    # Keep current directory relative to repository root.
    relative = Path(os.path.relpath(Path(".").resolve(), original_path))

    # Reroute standard streams
    original_mapped_std = get_mapped_std_streams(original_repository.all_files)
    mapped_std = {}
    for name, stream in original_mapped_std.items():
        stream_path = Path(path) / (Path(stream).relative_to(original_path))
        stream_path = stream_path.absolute()

        if not stream_path.exists():
            stream_path.parent.mkdir(parents=True, exist_ok=True)
            stream_path.touch()

        mapped_std[name] = stream_path

    _clean_streams(original_repository, original_mapped_std)

    new_cwd = Path(path) / relative
    new_cwd.mkdir(parents=True, exist_ok=True)

    isolation = Isolation(cwd=str(new_cwd), **mapped_std)
    isolation.__enter__()

    return repository, isolation, path, branch_name


def finalize_worktree(isolation, path, branch_name, delete, new_branch, merge_args=("--ff-only",), exception=None):
    """Cleanup and merge a previously created Git worktree."""
    exc_info: Tuple[Optional[Type[Any]], Any, Any] = (None, None, None)

    if exception:
        exc_info = (type(exception), exception, exception.__traceback__)

    isolation.__exit__(*exc_info)

    project_context.pop_context()

    repository = project_context.repository

    try:
        repository.run_git_command("merge", branch_name, *merge_args)
    except errors.GitCommandError:
        raise errors.FailedMerge(repository, branch_name, merge_args)

    if delete:
        repository.remove_worktree(path)

        if new_branch:
            # delete the created temporary branch
            repository.branches.remove(branch_name)

    if project_context.external_storage_requested:
        checkout_paths_from_storage()


def get_mapped_std_streams(lookup_paths, streams=("stdin", "stdout", "stderr")):
    """Get a mapping of standard streams to given paths."""
    # FIXME add device number too
    standard_inos = {}
    for stream in streams:
        try:
            stream_stat = os.fstat(getattr(sys, stream).fileno())
            key = stream_stat.st_dev, stream_stat.st_ino
            standard_inos[key] = stream
        except Exception:  # FIXME UnsupportedOperation
            pass
        # FIXME if not getattr(sys, stream).istty()

    def stream_inos(paths):
        """Yield tuples with stats and path."""
        for path in paths:
            try:
                stat = os.stat(path)
                key = (stat.st_dev, stat.st_ino)
                if key in standard_inos:
                    yield standard_inos[key], path
            except FileNotFoundError:  # pragma: no cover
                pass

        return []

    return dict(stream_inos(lookup_paths)) if standard_inos else {}


def _clean_streams(repository, mapped_streams):
    """Clean mapped standard streams."""
    for stream_name in ("stdout", "stderr"):
        stream = mapped_streams.get(stream_name)
        if not stream:
            continue

        absolute_path = get_absolute_path(stream, repository.path)
        path = os.path.relpath(absolute_path, start=repository.path)
        if path not in repository.files:
            os.remove(absolute_path)
        else:
            checksum = repository.get_object_hash(path=absolute_path, revision="HEAD")
            repository.copy_content_to_file(path=absolute_path, checksum=checksum, output_path=path)


def ensure_clean(ignore_std_streams=False):
    """Make sure the repository is clean."""
    repository = project_context.repository

    dirty_paths = get_dirty_paths(repository)
    mapped_streams = get_mapped_std_streams(dirty_paths)

    if ignore_std_streams:
        if dirty_paths - set(mapped_streams.values()):
            _clean_streams(repository, mapped_streams)
            raise errors.DirtyRepository(repository)
    elif repository.is_dirty(untracked_files=False):
        _clean_streams(repository, mapped_streams)
        raise errors.DirtyRepository(repository)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pointer file business logic."""

import os
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Tuple, Union, cast

from renku.core import errors
from renku.core.util.os import is_subpath
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Repository

if TYPE_CHECKING:
    from renku.domain_model.dataset import DatasetFile


def create_pointer_file(target: Union[str, Path], checksum: Optional[str] = None):
    """Create a new pointer file."""
    target = Path(target).resolve()

    if checksum is None:
        checksum = Repository.hash_object(target)
        assert checksum is not None, f"Cannot calculate checksum for '{target}'"

    while True:
        filename = f"{uuid.uuid4()}-{checksum}"
        path = project_context.pointers_path / filename
        if not path.exists():
            break

    # NOTE: If target is within the repo, add it as a relative symlink
    is_within_repo = is_subpath(target, base=project_context.path)
    source = cast(Union[str, bytes, Path], os.path.relpath(target, path.parent) if is_within_repo else target)

    try:
        os.symlink(source, path)
    except FileNotFoundError:
        raise errors.ExternalFileNotFound(target)

    return path


def is_linked_file_updated(path: Union[Path, str]) -> Tuple[bool, str]:
    """Check if an update to a linked file is available."""
    pointer_file = get_pointer_file(path=path)

    try:
        target = pointer_file.resolve(strict=True)
    except FileNotFoundError:
        target = pointer_file.resolve()
        raise errors.ExternalFileNotFound(target)

    new_checksum = Repository.hash_object(target)
    old_checksum = pointer_file.name.split("-")[-1]

    if new_checksum is None:
        raise errors.ExternalFileNotFound(target)

    updated = new_checksum != old_checksum

    return updated, new_checksum


def update_linked_file(path: Union[Path, str], checksum: Optional[str]):
    """Delete existing linked file and create a new one."""
    pointer_file = get_pointer_file(path=path)
    target = pointer_file.resolve()

    os.remove(pointer_file)
    absolute_path = project_context.path / path
    os.remove(absolute_path)

    create_external_file(target=target, path=absolute_path, checksum=checksum)


def create_external_file(target: Path, path: Union[Path, str], checksum: Optional[str] = None):
    """Create a new external file."""
    try:
        pointer_file = create_pointer_file(target=target, checksum=checksum)
        relative = os.path.relpath(pointer_file, Path(path).parent)
        os.symlink(relative, path)
    except OSError as e:
        raise errors.OperationError("Could not create symbolic link") from e


def get_pointer_file(path: Union[Path, str]) -> Path:
    """Return pointer file from an external file."""
    absolute_path = project_context.path / path
    link = absolute_path.parent / os.readlink(absolute_path)
    return project_context.path / link


def delete_external_file(dataset_file: "DatasetFile"):
    """Delete an external file."""
    if not dataset_file.is_external or not dataset_file.linked:
        return

    try:
        pointer_file = get_pointer_file(dataset_file.entity.path)
        if os.path.lexists(pointer_file):
            pointer_file.unlink()

        path = project_context.path / dataset_file.entity.path
        if os.path.lexists(path):
            path.unlink()
    except OSError as e:
        raise errors.InvalidFileOperation(f"Cannot remove file '{dataset_file.entity.path}'") from e



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parameters JSON-LD schemes."""

from marshmallow import EXCLUDE

from renku.command.schema.calamus import JsonLDSchema, Nested, fields, prov, renku, schema
from renku.domain_model.workflow.parameter import (
    CommandInput,
    CommandOutput,
    CommandParameter,
    CommandParameterBase,
    MappedIOStream,
    ParameterLink,
    ParameterMapping,
)


class MappedIOStreamSchema(JsonLDSchema):
    """MappedIOStream schema."""

    class Meta:
        """Meta class."""

        rdf_type = renku.IOStream
        model = MappedIOStream
        unknown = EXCLUDE

    id = fields.Id()
    stream_type = fields.String(renku.streamType)


class CommandParameterBaseSchema(JsonLDSchema):
    """CommandParameterBase schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandParameterBase, schema.Property]
        model = CommandParameterBase
        unknown = EXCLUDE

    default_value = fields.Raw(schema.defaultValue, load_default=None)
    description = fields.String(schema.description, load_default=None)
    id = fields.Id()
    name = fields.String(schema.name, load_default=None)
    position = fields.Integer(renku.position, load_default=None)
    prefix = fields.String(renku.prefix, load_default=None)
    derived_from = fields.String(prov.wasDerivedFrom, load_default=None)


class CommandParameterSchema(CommandParameterBaseSchema):
    """CommandParameter schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandParameter]
        model = CommandParameter
        unknown = EXCLUDE


class CommandInputSchema(CommandParameterBaseSchema):
    """CommandInput schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandInput]
        model = CommandInput
        unknown = EXCLUDE

    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, load_default=None)
    encoding_format = fields.List(schema.encodingFormat, fields.String(), load_default=None)


class CommandOutputSchema(CommandParameterBaseSchema):
    """CommandOutput schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandOutput]
        model = CommandOutput
        unknown = EXCLUDE

    create_folder = fields.Boolean(renku.createFolder, load_default=False)
    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, load_default=None)
    encoding_format = fields.List(schema.encodingFormat, fields.String(), load_default=None)


class ParameterMappingSchema(CommandParameterBaseSchema):
    """ParameterMapping schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.ParameterMapping]
        model = ParameterMapping
        unknown = EXCLUDE

    mapped_parameters = Nested(
        renku.mapsTo,
        ["ParameterMappingSchema", CommandInputSchema, CommandOutputSchema, CommandParameterSchema],
        many=True,
    )


class ParameterLinkSchema(JsonLDSchema):
    """ParameterLink schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.ParameterLink, schema.Property]
        model = ParameterLink
        unknown = EXCLUDE

    id = fields.Id()
    source = fields.Nested(renku.linkSource, [CommandOutputSchema])
    sinks = fields.Nested(renku.linkSink, [CommandInputSchema, CommandParameterSchema], many=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template serializers."""

from urllib.parse import urlparse

from marshmallow import Schema, ValidationError, fields, post_load, pre_load
from yagup import GitURL
from yagup.exceptions import InvalidURL

from renku.core.util.os import normalize_to_ascii
from renku.domain_model.dataset import ImageObjectRequestJson
from renku.ui.service.config import TEMPLATE_CLONE_DEPTH_DEFAULT
from renku.ui.service.serializers.cache import ProjectCloneContext, RepositoryCloneRequest
from renku.ui.service.serializers.rpc import JsonRPCResponse
from renku.ui.service.serializers.templates import TemplateParameterSchema
from renku.ui.service.utils import normalize_git_url


class ManifestTemplateSchema_1_5(Schema):
    """Manifest template schema."""

    description = fields.String(required=True, metadata={"description": "Description of the template"})
    folder = fields.String(required=True, metadata={"description": "Folder the template resides in"})
    name = fields.String(required=True, metadata={"description": "Name of the template"})
    variables = fields.Dict(
        load_default={}, metadata={"description": "Dictionary of values that can be set on this template"}
    )
    icon = fields.String(
        load_default=None, metadata={"description": "base64 encoded icon for the template in PNG format"}
    )


class ManifestTemplatesResponse_1_5(Schema):
    """Manifest templates response."""

    templates = fields.List(fields.Nested(ManifestTemplateSchema_1_5), required=True)


class ManifestTemplatesResponseRPC_1_5(JsonRPCResponse):
    """RPC schema for listing manifest templates."""

    result = fields.Nested(ManifestTemplatesResponse_1_5)


class ManifestTemplatesRequest_v2_2(RepositoryCloneRequest):
    """Request schema for listing manifest templates."""

    url = fields.String(required=True)
    depth = fields.Integer(load_default=TEMPLATE_CLONE_DEPTH_DEFAULT)

    @pre_load()
    def set_git_url(self, data, **kwargs):
        """Set git_url field."""
        if data.get("url"):
            data["git_url"] = data.get("url")
        data["template_git_url"] = data["git_url"]
        return data


class ProjectTemplateRequest_v2_2(ProjectCloneContext, ManifestTemplatesRequest_v2_2):
    """Request schema for listing manifest templates."""

    identifier = fields.String(required=True, metadata={"description": "Indentifier of the template"})
    initial_branch = fields.String(
        load_default=None, metadata={"description": "Name for the initial branch in the new project."}
    )
    parameters = fields.List(
        fields.Nested(TemplateParameterSchema), load_default=[], metadata={"description": "Template parameters"}
    )
    project_name = fields.String(required=True, metadata={"description": "Project name"})
    project_namespace = fields.String(required=True, metadata={"description": "Project namespace"})
    project_repository = fields.String(required=True, metadata={"description": "Project remote repository"})
    project_description = fields.String(load_default=None, metadata={"description": "Project description"})
    project_keywords = fields.List(fields.String(), load_default=None, metadata={"description": "Project keywords"})
    project_custom_metadata = fields.Dict(
        load_default=None, metadata={"description": "Project custom JSON-LD metadata"}
    )
    data_directory = fields.String(
        load_default=None, metadata={"description": "Base dataset data directory in project. Defaults to 'data/'"}
    )
    image = fields.Nested(ImageObjectRequestJson, load_default=None)

    @post_load()
    def add_required_fields(self, data, **kwargs):
        """Add necessary fields."""
        data["template_git_url"] = data.get("git_url")
        project_name_stripped = normalize_to_ascii(data["project_name"])
        project_name_stripped = normalize_git_url(project_name_stripped)
        if len(project_name_stripped) == 0:
            raise ValidationError("Project name contains only unsupported characters")
        new_project_url = f"{data['project_repository']}/{data['project_namespace']}/{project_name_stripped}"
        try:
            _ = GitURL.parse(new_project_url)
        except InvalidURL as e:
            raise ValidationError("`git_url` contains unsupported characters") from e

        project_slug = f"{data['project_namespace']}/{project_name_stripped}"
        data["new_project_url"] = new_project_url
        data["project_name_stripped"] = project_name_stripped
        data["project_slug"] = project_slug

        new_project_url_parsed = urlparse(new_project_url)
        url = "oauth2:{}@{}".format(data["token"], new_project_url_parsed.netloc)
        data["new_project_url_with_auth"] = new_project_url_parsed._replace(netloc=url).geturl()

        return data



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service user cache management."""
from renku.ui.service.cache.base import BaseCache
from renku.ui.service.cache.models.user import User
from renku.ui.service.cache.serializers.user import UserSchema


class UserManagementCache(BaseCache):
    """User management cache."""

    user_schema = UserSchema()

    def ensure_user(self, user_data):
        """Ensure user data registered in a cache."""
        user_obj: User = self.user_schema.load(user_data)

        try:
            user_obj = User.get(User.user_id == user_obj.user_id and User.token == user_obj.token)
        except ValueError:
            user_obj.save()

        return user_obj

    @staticmethod
    def get_user(user_id):
        """Get specific user."""
        try:
            user_obj = User.get(User.user_id == user_id)
        except ValueError:
            return

        return user_obj

    @staticmethod
    def get_users():
        """Get all users."""
        return User.all()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check for large files in Git history."""

from renku.command.util import WARNING
from renku.core.storage import check_external_storage, check_lfs_migrate_info


def check_lfs_info(**_):
    """Checks if files in history should be in LFS.

    Args:
        _: keyword arguments.

    Returns:
        Tuple of whether project structure is valid, if an automated fix is available and string of found problems.
    """
    if not check_external_storage():
        return True, False, None

    files = check_lfs_migrate_info()

    if not files:
        return True, False, None

    message = (
        WARNING
        + "Git history contains large files - consider moving them "
        + "to external storage like git LFS using 'renku storage migrate'\n\t"
        + "\n\t".join(files)
        + "\n"
    )

    return False, False, message



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku storage command."""

from typing import List

from pydantic import ConfigDict, validate_call

from renku.command.command_builder.command import Command
from renku.core.storage import (
    check_lfs_migrate_info,
    check_requires_tracking,
    clean_storage_cache,
    migrate_files_to_lfs,
    pull_paths_from_storage,
)
from renku.core.util import communication
from renku.domain_model.project_context import project_context


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _check_lfs(everything: bool = False):
    """Check if large files are not in lfs.

    Args:
        everything: Whether to check whole history (Default value = False).

    Returns:
        List of large files.
    """
    files = check_lfs_migrate_info(everything)

    if files:
        communication.warn("Git history contains large files\n\t" + "\n\t".join(files))

    return files


def check_lfs_command():
    """Check lfs command."""
    return Command().command(_check_lfs)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _fix_lfs(paths: List[str]):
    """Migrate large files into lfs.

    Args:
        paths(List[str]): Paths to migrate to LFS.
    """
    migrate_files_to_lfs(paths)


def fix_lfs_command():
    """Fix lfs command."""
    return (
        Command()
        .command(_fix_lfs)
        .require_clean()
        .require_migration()
        .with_database(write=True)
        .with_commit(commit_if_empty=False)
    )


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _pull(paths: List[str]):
    """Pull the specified paths from external storage.

    Args:
        paths(List[str]): Paths to pull from LFS.
    """
    pull_paths_from_storage(project_context.repository, *paths)


def pull_command():
    """Command to pull the specified paths from external storage."""
    return Command().command(_pull)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _clean(paths: List[str]):
    """Remove files from lfs cache/turn them back into pointer files.

    Args:
        paths:List[str]: Paths to turn back to pointer files.
    """
    untracked_paths, local_only_paths = clean_storage_cache(*paths)

    if untracked_paths:
        communication.warn(
            "These paths were ignored as they are not tracked"
            + " in git LFS:\n\t{}\n".format("\n\t".join(untracked_paths))
        )

    if local_only_paths:
        communication.warn(
            "These paths were ignored as they are not pushed to "
            + "a remote with git LFS:\n\t{}\n".format("\n\t".join(local_only_paths))
        )


def clean_command():
    """Command to remove files from lfs cache/turn them back into pointer files."""
    return Command().command(_clean)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _check_lfs_hook(paths: List[str]):
    """Check if paths should be in LFS.

    Args:
        paths(List[str]): Paths to check

    Returns:
        List of files that should be in LFS.
    """
    return check_requires_tracking(*paths)


def check_lfs_hook_command():
    """Command to pull the specified paths from external storage."""
    return Command().command(_check_lfs_hook)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repository cache interface."""

import os
import shutil
import uuid
from datetime import datetime
from typing import Optional
from urllib.parse import urlparse

import portalocker
from marshmallow import ValidationError

from renku.command.clone import project_clone_command
from renku.core import errors
from renku.core.util.contexts import renku_project_context
from renku.core.util.os import normalize_to_ascii
from renku.domain_model.git import GitURL
from renku.infrastructure.repository import Repository
from renku.ui.service.cache import ServiceCache
from renku.ui.service.cache.models.project import Project
from renku.ui.service.cache.models.user import User
from renku.ui.service.config import PROJECT_CLONE_DEPTH_DEFAULT
from renku.ui.service.errors import IntermittentCacheError, IntermittentLockError
from renku.ui.service.interfaces.repository_cache import IRepositoryCache
from renku.ui.service.logger import service_log
from renku.ui.service.utils import normalize_git_url


class LocalRepositoryCache(IRepositoryCache):
    """Cache for project repos stored on local disk."""

    def get(
        self,
        cache: ServiceCache,
        git_url: str,
        branch: Optional[str],
        user: User,
        shallow: bool = True,
        commit_sha: Optional[str] = None,
    ) -> Project:
        """Get a project from cache (clone if necessary)."""
        if git_url is None:
            raise ValidationError("Invalid `git_url`, URL is empty", "git_url")
        # Note: walrus turns None into empty strings
        branch = branch or ""
        commit_sha = commit_sha or ""

        git_url = normalize_git_url(git_url)
        try:
            project = Project.get(
                (Project.user_id == user.user_id)
                & (Project.git_url == git_url)
                & (Project.branch == branch)
                & (Project.commit_sha == commit_sha)
            )
        except ValueError:
            # project not found in DB
            return self._clone_project(cache, git_url, branch, user, shallow, commit_sha)

        if not project.abs_path.exists():
            # cache folder doesn't exist anymore
            project.delete()
            return self._clone_project(cache, git_url, branch, user, shallow, commit_sha)

        if not shallow and project.is_shallow:
            self._unshallow_project(project, user)

        self._maybe_update_cache(project, user)

        if not project.initialized:
            raise errors.UninitializedProject(project.git_url)

        self._update_project_access_date(project)

        return project

    def evict(self, project: Project):
        """Evict a project from cache."""
        try:
            with project.write_lock():
                service_log.debug(f"purging project {project.project_id}:{project.name}")
                project.purge()
        except FileNotFoundError:
            project.delete()
        except Exception as e:
            service_log.error(f"Couldn't purge project {project.project_id}:{project.name} from cache", exc_info=e)

    def evict_expired(self):
        """Evict expired projects from cache."""
        for project in Project.all():
            if project.ttl_expired():
                self.evict(project)

    def _update_project_access_date(self, project: Project):
        """Update the access date of the project to current datetime."""
        project.accessed_at = datetime.utcnow()
        project.save()

    def _clone_project(
        self,
        cache: ServiceCache,
        git_url: str,
        branch: Optional[str],
        user: User,
        shallow: bool = True,
        commit_sha: Optional[str] = None,
    ) -> Project:
        """Clone a project to cache."""
        git_url = normalize_git_url(git_url)

        try:
            parsed_git_url = GitURL.parse(git_url)
        except UnicodeError as e:
            raise ValidationError("`git_url` contains unsupported characters", "git_url") from e
        except errors.InvalidGitURL as e:
            raise ValidationError("Invalid `git_url`", "git_url") from e

        if parsed_git_url.owner is None or parsed_git_url.name is None:
            raise ValidationError("Invalid `git_url`, missing owner or repository", "git_url")
        if branch == "":
            branch = None
        if commit_sha == "":
            commit_sha = None

        project_data = {
            "project_id": uuid.uuid4().hex,
            "owner": parsed_git_url.owner,
            "name": parsed_git_url.name,
            "slug": normalize_to_ascii(parsed_git_url.name),
            "clone_depth": PROJECT_CLONE_DEPTH_DEFAULT if shallow else None,
            "branch": branch,
            "git_url": git_url,
            "user_id": user.user_id,
            "commit_sha": commit_sha,
        }
        project = cache.make_project(user, project_data, persist=False)

        # NOTE: Create parent dir so lock file can be created.
        project.abs_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            with project.write_lock(), renku_project_context(project.abs_path, check_git_path=False):
                try:
                    # NOTE: If two requests ran at the same time, by the time we acquire the lock a project might
                    # already be cloned by an earlier request.
                    found_project = Project.get(
                        (Project.user_id == user.user_id)
                        & (Project.git_url == git_url)
                        & (Project.branch == branch)
                        & (Project.commit_sha == commit_sha)
                        & (Project.project_id != project.project_id)
                    )
                except ValueError:
                    pass
                else:
                    if found_project.abs_path.exists():
                        service_log.debug(f"project already cloned, skipping clone: {git_url}")
                        self._update_project_access_date(found_project)
                        return found_project

                # clean directory in case of previous failed state
                # NOTE: we only want to delete the contents, NOT the folder itself, in case it's still referenced
                for root, dirs, files in os.walk(project.abs_path):
                    for f in files:
                        os.unlink(os.path.join(root, f))
                    for d in dirs:
                        shutil.rmtree(os.path.join(root, d))

                repo, project.initialized = (
                    project_clone_command()
                    .build()
                    .execute(
                        git_url_with_auth(project, user),
                        path=project.abs_path,
                        depth=project.clone_depth,
                        raise_git_except=True,
                        config={
                            "user.name": user.fullname,
                            "user.email": user.email,
                            "pull.rebase": False,
                        },
                        checkout_revision=commit_sha or project.branch,
                    )
                ).output
                project.save()

                service_log.debug(f"project successfully cloned: {repo}")

                if not project.initialized:
                    raise errors.UninitializedProject(project.git_url)

                return project
        except (portalocker.LockException, portalocker.AlreadyLocked, errors.LockError) as e:
            raise IntermittentLockError() from e

    def _unshallow_project(self, project: Project, user: User):
        """Turn a shallow clone into a full clone."""
        if project.commit_sha is not None:
            # NOTE: A project in a detached head state at a specific commit SHA does not make sense to be unshallowed
            return
        try:
            with project.write_lock(), Repository(project.abs_path) as repository:
                try:
                    # NOTE: It could happen that repository is already un-shallowed,
                    # in this case we don't want to leak git exception, but still want to fetch.
                    repository.fetch("origin", repository.active_branch, unshallow=True)
                except errors.GitCommandError:
                    repository.fetch("origin", repository.active_branch)

                repository.reset(f"origin/{repository.active_branch}", hard=True)
                project.clone_depth = None
                project.save()
        except (portalocker.LockException, portalocker.AlreadyLocked, errors.LockError) as e:
            raise IntermittentLockError() from e

    def _maybe_update_cache(self, project: Project, user: User):
        """Update the cache from the remote if it's out of date."""
        from renku.ui.service.controllers.api.mixins import PROJECT_FETCH_TIME

        if project.fetch_age < PROJECT_FETCH_TIME:
            return

        if project.commit_sha is not None and project.commit_sha != "":
            # NOTE: A project in a detached head state at a specific commit SHA cannot be updated
            return

        try:
            with project.write_lock(), Repository(project.abs_path) as repository:
                try:
                    # NOTE: it rarely happens that origin is not reachable. Try again if it fails.
                    repository.fetch(
                        "origin",
                        repository.active_branch,
                        depth=project.clone_depth
                        if project.clone_depth is not None and project.clone_depth > 0
                        else None,
                    )
                    repository.reset(f"origin/{repository.active_branch}", hard=True)
                except errors.GitCommandError as e:
                    project.purge()
                    raise IntermittentCacheError(e)

                project.last_fetched_at = datetime.utcnow()
                project.save()
        except (portalocker.LockException, portalocker.AlreadyLocked, errors.LockError) as e:
            raise IntermittentLockError() from e


def git_url_with_auth(project: Project, user: User):
    """Format url with auth."""
    git_url = urlparse(normalize_git_url(project.git_url))

    url = "oauth2:{}@{}".format(user.token, git_url.netloc)
    return git_url._replace(netloc=url).geturl()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache job related models."""
from walrus import DateTimeField, JSONField, Model, SetField, TextField

from renku.ui.service.cache.base import BaseCache

# User job states
USER_JOB_STATE_ENQUEUED = "ENQUEUED"
USER_JOB_STATE_IN_PROGRESS = "IN_PROGRESS"
USER_JOB_STATE_COMPLETED = "COMPLETED"
USER_JOB_STATE_FAILED = "FAILED"


class Job(Model):
    """Job cache model."""

    __database__ = BaseCache.model_db
    __namespace__ = BaseCache.namespace

    created_at = DateTimeField()
    updated_at = DateTimeField()

    job_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)
    project_id = TextField(index=True)

    renku_op = TextField()
    state = TextField()
    extras = JSONField()
    client_extras = TextField()
    branch = TextField()

    ctrl_context = JSONField()
    ctrl_result = JSONField()

    locked = SetField()

    def in_progress(self):
        """Mark job in progress."""
        self.state = USER_JOB_STATE_IN_PROGRESS
        self.save()

    def complete(self):
        """Mark job as completed."""
        self.state = USER_JOB_STATE_COMPLETED
        self.save()

    def fail_job(self, error):
        """Mark job as failed."""
        self.state = USER_JOB_STATE_FAILED

        if not self.extras:
            self.extras = {}

        self.extras["error"] = error
        self.save()

    def update_extras(self, key, value):
        """Update extras field."""
        if not self.extras:
            self.extras = {key: value}
        else:
            self.extras[key] = value
